"""Classes for working with topology of tilings.

Together the `Topology`, `Tile`, `Vertex`, `Edge`, and
`weavingspace.symmetry.Symmetries` classes enable extraction of the topological
structure of periodic `weavingspace.tileable.Tileable` objects so that
modification of equivalent tiles can be carried out while retaining
tileability.

It is important to note that these are not fully generalised classes and
methods, that is, the Topology object that is supported is not a permanent
'backing' data structure for our Tileable objects. While it might become that
in time, it is not yet such a data structure. Instead usage is

    tile = TileUnit(...)
    topology = Topology(tile)
    topology = topology.transform_*(...)
    new_tile = topology.tile_unit

A Topology plot function is provided for a user to be able to see what they are
doing, because how edges and vertices in a tiling are labelled under tile
equivalences is an essential step in the process.

Note that these classes do not precisely represent the distinctions made in the
mathematical literature between tiling vertices and tile corners, or between
tiling edges and tile sides.
"""
from __future__ import annotations

import inspect
import itertools
import pickle
import string
from collections import defaultdict
from typing import TYPE_CHECKING

import geopandas as gpd
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import shapely.affinity as affine
import shapely.geometry as geom
from scipy import interpolate

from weavingspace import (
  ShapeMatcher,
  Symmetries,
  Tileable,
  Transform,
  tiling_utils,
)

if TYPE_CHECKING:
  from collections.abc import Callable, Iterable

# all two letter combinations of the alphabet for labelling
LABELS = \
  list(string.ascii_letters.upper()) + \
  ["".join(x) for x in
   itertools.combinations(list(string.ascii_letters.upper()), 2)]
labels = \
  list(string.ascii_letters.lower()) + \
  ["".join(x) for x in
   itertools.combinations(list(string.ascii_letters.lower()), 2)]


class Topology:
  """Class to represent topology of a Tileable object.

  NOTE: It is important that get_local_patch return the tileable elements and
  the translated copies in consistent sequence, i.e. if there are (say) four
  tiles in the unit, the local patch should be 1 2 3 4 1 2 3 4 1 2 3 4 ... and
  so on. This is because self.tiles[i % n_tiles] is frequently used to
  reference the base unit Tile which corresponds to self.tiles[i].
  """

  tileable: Tileable
  """the Tileable on which the topology will be based."""
  tiles: list[Tile]
  """list of the Tiles in the topology. We use polygons returned by the
  tileable.get_local_patch method for these. That is the base tiles and 8
  adjacent copies (for a rectangular tiling), or 6 adjacent copies (for a
  hexagonal tiling)."""
  points: dict[int, Vertex]
  """dictionary of all points (vertices and corners) in the tiling, keyed by
  Vertex ID."""
  edges: dict[tuple[int,int], Edge]
  """dictionary of the tiling edges, keyed by Edge ID."""
  unique_tile_shapes: list[geom.Polygon]
  """a 'reference' tile shape one per tile shape (up to vertices, so two tiles
  might be the same shape, but one might have extra vertices induced by the
  tiling and hence is a different shape under this definition)."""
  dual_tiles: list[geom.Polygon]
  """list of geom.Polygons from which a dual tiling might be constructed."""
  n_tiles: int = 0
  """number of tiles in the base Tileable (retained for convenience)."""
  shape_groups: list[list[int]]
  """list of lists of tile IDs distinguished by shape and optionally tile_id"""
  tile_matching_transforms: list[tuple[float]]
  """shapely transform tuples that map tiles onto other tiles"""
  tile_transitivity_classes: list[tuple[int]]
  """list of lists of tile IDs in each transitivity class"""
  vertex_transitivity_classes: list[list[int]]
  """list of lists of vertex IDs in each transitivity class"""
  edge_transitivity_classes: list[list[tuple[int]]]
  """list of lists of edge IDs in each transitivity class"""

  def __init__(
      self,
      unit: Tileable,
      ignore_tile_ids:bool = True,
    ) -> None:
    """Class constructor.

    Args:
      unit (Tileable): the Tileable whose topology is required.
      ignore_tile_ids (bool): (EXPERIMENTAL) if True then only consider the tile
        shapes, not labels. If False consider any labels. Defaults to True.

    """
    # Note that the order of these setup steps is fragile sometimes
    # not obviously so.
    self.tileable = unit # keep this for reference
    self.n_tiles = self.tileable.tiles.shape[0]
    self._initialise_points_into_tiles()
    self._setup_vertex_tile_relations()
    self._setup_edges()
    self._copy_base_tiles_to_patch()
    self._assign_vertex_and_edge_base_IDs()
    self._identify_distinct_tile_shapes(ignore_tile_ids)
    self._find_tile_transitivity_classes(ignore_tile_ids)
    self._find_vertex_transitivity_classes(ignore_tile_ids)
    self._find_edge_transitivity_classes(ignore_tile_ids)
    self.generate_dual()


  def __str__(self) -> str:
    """Return string representation of this Topology.

    Returns:
      str: a message that recommends examining the tiles, points and edges
        attributes.

    """
    return (f"""Topology of Tileable with {self.n_tiles} tiles.\n
            Examine .tiles, .points and .edges for more details.""")


  def __repr__(self) -> str:
    return str(self)


  def _initialise_points_into_tiles(self, debug:bool = False) -> None:
    """Set up dictionary of unique point locations and assign them to Tiles.

    Args:
      debug (bool): if True prints useful debugging information.

    """
    shapes = self.tileable.get_local_patch(r = 1, include_0 = True).geometry
    shapes = [tiling_utils.get_clean_polygon(s) for s in shapes]
    self.n_1_order_patch = len(shapes)
    labels = list(self.tileable.tiles.tile_id) * (len(shapes) // self.n_tiles)
    self.tiles = []
    self.points = {}
    for (i, shape), label in zip(enumerate(shapes), labels, strict = True):
      tile = Tile(i)
      tile.label = label
      tile.base_ID = tile.ID % self.n_tiles
      self.tiles.append(tile)
      tile.corners = []
      corners = tiling_utils.get_corners(shape, repeat_first = False)
      for c in corners:
        prev_vertex = None
        for p in reversed(self.points.values()):
          if c.distance(p.point) <= 2 * tiling_utils.RESOLUTION:
            # found an already existing vertex, so add to the tile and break
            prev_vertex = p
            tile.corners.append(prev_vertex)
            break
        if prev_vertex is None:
          # new vertex, add it to the topology dictionary and to the tile
          v = self.add_vertex(c)
          tile.corners.append(v)
          if debug:
            print(f"Added new Vertex {v} to Tile {i}")


  def _setup_vertex_tile_relations(self, debug:bool = False) -> None:
    """Determine relations between vertices and tiles.

    In particular vertices along tile edges that are not yet included in their
    list of vertices are added. Meanwhile vertex lists of incident tiles and
    neighbours are set up.

    Args:
      debug (bool): if True prints debugging information.

    """
    # we do this for all tiles in the radius-1 local patch
    for tile in self.tiles:
      if debug:
        print(f"Checking for vertices incident on Tile {tile.ID}")
      corners = []
      # performance is much improved using the vertex IDs
      initial_corner_IDs = tile.get_corner_IDs()
      # we need the current shape (not yet set) to check for incident vertices
      shape = geom.Polygon([c.point for c in tile.corners])
      # get points incident on tile boundary, not already in the tile corners
      new_points = [v for v in self.points.values()
                    if v.ID not in initial_corner_IDs and
                    v.point.distance(shape) <= 2 * tiling_utils.RESOLUTION]
      # iterate over sides of the tile to see which the vertex is incident on
      for c1, c2 in zip(tile.corners, tile.corners[1:] + tile.corners[:1],
                        strict = True):
        all_points = [c1, c2]
        if len(new_points) > 0:
          if debug:
            print(f"{[v.ID for v in new_points]} incident on tile")
          ls = geom.LineString([c1.point, c2.point])
          to_insert = [v for v in new_points
                       if v.point.distance(ls) <= 2 * tiling_utils.RESOLUTION]
          if len(to_insert) > 0:
            # sort by distance along the side
            d_along = sorted([(ls.line_locate_point(v.point), v)
                              for v in to_insert], key = lambda x: x[0])
            to_insert = [v for d, v in d_along]
            all_points = all_points[:1] + to_insert + all_points[1:]
        corners.extend(all_points[:-1])
        for x1, x2 in zip(all_points[:-1], all_points[1:], strict = True):
          # x2 will add the tile and neigbour when we get to the next side
          # every vertex gets a turn!
          x1.add_tile(tile)
          x1.add_neighbour(x2.ID)
      tile.corners = corners
      tile.set_shape_from_corners()


  def _setup_edges(
      self,
      debug:bool = False,
    ) -> None:
    """Set up the tiling edges.

    First vertices in the base tiles are classified as tiling vertices or not -
    only these can be classified reliably (e.g vertices on the perimeter are
    tricky).. Up to here all vertices have been considered tiling vertices.

    Second edges are created by traversing tile corner lists. Edges are stored
    once only by checking for edges in the reverse direction already in the
    edges dictionary. Edge right and left tiles are initialised.

    Third tile edge direction lists are initialised.

    Args:
      debug (bool): if True print debug messages. Defaults to False.

    """
    # classify vertices in the base tiles
    for tile in self.tiles[:self.n_tiles]:
      for v in tile.corners:
        v.is_tiling_vertex = len(v.neighbours) > 2
    if debug:
      print("Classified base tile vertices")
    self.edges = {}
    for tile in self.tiles:
      if debug:
        print(f"Adding edges from Tile {tile.ID}")
      tile.edges = []
      vertices = [v for v in tile.corners if v.is_tiling_vertex]
      # note that through here finding ints in lists is much faster than
      # finding Vertex objects, hence we use lists of IDs not Vertex objects
      if len(vertices) > 1:
        for v1, v2 in zip(vertices, vertices[1:] + vertices[:1], strict = True):
          corner_IDs = tile.get_corner_IDs()
          idx1 = corner_IDs.index(v1.ID)
          idx2 = corner_IDs.index(v2.ID)
          if idx1 < idx2:
            corners = corner_IDs[idx1:(idx2 + 1)]
          else:
            corners = corner_IDs[idx1:] + corner_IDs[:(idx2 + 1)]
          ID = (corners[0], corners[-1])
          if ID not in self.edges:
            # check that reverse direction edge is not present first
            r_ID = ID[::-1]
            if r_ID in self.edges:
              # if it is, then add it and set left_tile
              if debug:
                print(f"reverse edge {r_ID} found")
              e = self.edges[r_ID]
              e.left_tile = tile
              tile.edges.append(e)
            else:
              # we've found a new edge so make and add it
              if debug:
                print(f"adding new edge {corners}")
              e = self.add_edge([self.points[c] for c in corners])
              e.right_tile = tile
              tile.edges.append(e)
      # initialise the edge direction information in the tile
      tile.set_edge_directions()


  def _assign_vertex_and_edge_base_IDs(self) -> None:
    """Assign the base_ID attributes of vertices and edges.

    These allow us to determine corrspondences between vertices and edges in
    the 'base' tiles in the Topology tileable, and those we have added at
    radius 1 for labelling and visualisation.
    """
    self._assign_vertex_base_IDs()
    self._assign_edge_base_IDs()


  def _assign_vertex_base_IDs(self) -> None:
    """Assign base_ID attribute of vertices."""
    # assign vertex base_ID frome the core tiles
    for tile0 in self.tiles[:self.n_tiles]:
      for v in tile0.get_corner_IDs():
        self.points[v].base_ID = v
    # assign others from their corresponding vertex in the core
    for t0 in self.tiles[:self.n_tiles]:
      for t1 in self.tiles[self.n_tiles:]:
        if t1.base_ID == t0.base_ID:
          for v0, v1 in zip(t0.corners, t1.corners, strict = True):
            v1.base_ID = v0.base_ID


  def _assign_edge_base_IDs(self) -> None:
    """Assign base_ID attribute of edges, based on base_IDs of vertices."""
    for tile0 in self.tiles[:self.n_tiles]:
      for e in tile0.get_edge_IDs():
        self.edges[e].base_ID = e
    for t0 in self.tiles[:self.n_tiles]:
      for t1 in self.tiles[self.n_tiles:]:
        if t1.base_ID == t0.base_ID:
          for e0, e1 in zip(t0.edges, t1.edges, strict = True):
            e1.base_ID = e0.base_ID


  def _copy_base_tiles_to_patch(self) -> None:
    """Copy attributes of base tiles to corresponding tiles in radius-1 patch.

    This requires:

    First, inserting any additional corners in the base tiles not found in the
    radius-1 tiles.

    Second, any vertices in the base tiles that are NOT tiling vertices are
    applied to radius-1 tiles leading to merging of some edges.
    """
    # the number of tiles in the base + radius-1
    n_r1 = len(self.tiles)
    # first add any missing vertices to the non-base tiles
    # we add all the missing vertices before doing any merges
    for base in self.tiles[:self.n_tiles]:
      for other in self.tiles[base.ID:n_r1:self.n_tiles]:
        self._match_reference_tile_vertices(base, other)
    # then merge any edges that meet at a corner
    for base in self.tiles[:self.n_tiles]:
      for other in self.tiles[base.ID:n_r1:self.n_tiles]:
        self._match_reference_tile_corners(base, other)


  def _match_reference_tile_vertices(
      self,
      tile1:Tile,
      tile2:Tile,
    ) -> None:
    """Add vertices to tile2 so it matches tile1 adjusting edges as required.

    This assumes the tiles are the same shape, but that tile2 may be missing
    some tiling vertices along some edges.

    Args:
      tile1 (Tile): reference tile.
      tile2 (Tile): tile to change.

    """
    to_add = len(tile1.corners) - len(tile2.corners)
    while to_add > 0:
      # find the reference x-y offset
      dxy = (tile2.centre.x - tile1.centre.x, tile2.centre.y - tile1.centre.y)
      for i, t1c in enumerate([c.point for c in tile1.corners]):
        t2c = tile2.corners[i % len(tile2.corners)].point
        if abs((t2c.x - t1c.x) - dxy[0]) > 10 * tiling_utils.RESOLUTION or \
           abs((t2c.y - t1c.y) - dxy[1]) > 10 * tiling_utils.RESOLUTION:
          # add vertex to t2 by copying the t1 vertex appropriately offset
          # note that this might alter the length of t2.corners
          v = self.add_vertex(geom.Point(t1c.x + dxy[0], t1c.y + dxy[1]))
          v.is_tiling_vertex = True
          old_edge, new_edges = tile2.insert_vertex_at(v, i)
          del self.edges[old_edge]
          for e in new_edges:
            self.edges[e.ID] = e
          to_add = to_add - 1


  def _match_reference_tile_corners(
      self,
      tile1:Tile,
      tile2:Tile,
    ) -> None:
    """Make vertices that are corners in tile1 corners in tile2.

    Edges are merged as required.

    Args:
        tile1 (Tile): reference tile.
        tile2 (Tile): tile to make match.

    """
    vs_to_change = [
      vj for vi, vj in zip(tile1.corners, tile2.corners, strict = True)
      if not vi.is_tiling_vertex and vj.is_tiling_vertex]
    if len(vs_to_change) > 0:
      for v in vs_to_change:
        v.is_tiling_vertex = False
        # it's a corner not an edge so will have no more than 2 v.tiles
        old_edges, new_edge = v.tiles[0].merge_edges_at_vertex(v)
        for e in old_edges:
          del self.edges[e]
        self.edges[new_edge.ID] = new_edge

  def _identify_distinct_tile_shapes(
      self,
      ignore_tile_id_labels:bool = True,
    ) -> None:
    """Identify unique tiles based on their symmetries and shapes.

    At the same time assembles a list of the affine transforms under which
    matches occurs since these are potential symmetries of the tiling.

    TODO: reimplement consideration of tile_id

    Args:
      ignore_tile_id_labels (bool): if True only the shape of tiles matters; if
        False the tile_id label is also considered. Defaults to True.

    """
    if ignore_tile_id_labels:
      matches = {}
      offsets = {}
      for tile in self.tiles[:self.n_tiles]:
        matches[tile.base_ID] = [tile.base_ID]
        matched = False
        s = Symmetries(tile.shape)
        for other in self.tiles[:self.n_tiles]:
          if other.ID > tile.ID:
            offset = s.get_corner_offset(other.shape)
            if offset is not None:
              offsets[tile.base_ID] = offset
              matches[tile.base_ID].append(other.base_ID)
              matched = True
        if not matched:
          offsets[tile.base_ID] = 0
      base_groups = list(
        nx.connected_components(nx.from_dict_of_lists(matches)))
      self.shape_groups = []
      for i, group in enumerate(base_groups):
        full_group = []
        for tile in self.tiles:
          if tile.base_ID in group:
            tile.shape_group = i
            tile.offset_corners(offsets[tile.base_ID])
            full_group.append(tile.ID)
        self.shape_groups.append(full_group)
    else:
      self.shape_groups = []
      for ti in self.tiles[:self.n_tiles]:
        self.shape_groups.append(
          [tj.ID for tj in self.tiles if tj.base_ID == ti.base_ID])
      for i, group in enumerate(self.shape_groups):
        for j in group:
          self.tiles[j].shape_group = i

  def _find_tile_transitivity_classes(
      self,
      ignore_tile_id_labels:bool = True,
    ) -> None:
    """Find tiles equivalent under symmetries.

    Also update the tile_matching_transforms attribute to contain only those
    transforms that pass this test.

    Args:
      ignore_tile_id_labels (bool): if True then consider only shapes; if False
        also consider labels of tiles. Defaults to True.

    """
    self.tile_matching_transforms = \
      self.get_potential_symmetries(ignore_tile_id_labels)
    if ignore_tile_id_labels:
      base_tiles = self.tiles[:self.n_tiles]
      # it is quicker (should be!) to only do within shape group tests
      # often there is only one when it will make no difference
      by_group_equivalent_tiles = []
      # maintain a set of transforms still potentially tiling symmetries
      poss_transforms = set(self.tile_matching_transforms.keys())
      # and a dictionary of booleans tracking which transforms are still valid
      eq_under_transform = dict.fromkeys(poss_transforms, True)
      for g, _ in enumerate(self.shape_groups):
        by_group_equivalent_tiles.append(set())
        source_tiles = [tile for tile in base_tiles if tile.shape_group == g]
        target_tiles = [tile for tile in self.tiles if tile.shape_group == g]
        for tr in poss_transforms:
          transform = self.tile_matching_transforms[tr].transform
          matched_tiles = {}
          eq_under_transform[tr] = True
          for source_tile in source_tiles:
            matched_tile_id = self._match_geoms_under_transform(
              source_tile, target_tiles, transform)
            if matched_tile_id == -1:
              eq_under_transform[tr] = False
              break
            matched_tiles[source_tile.ID] = matched_tile_id # actually a base_ID
          if eq_under_transform[tr]:
            for k, v in matched_tiles.items():
              # here we record the transform, in case it is later invalidated
              by_group_equivalent_tiles[g].add((tr, k, v))
        # remove valid transforms that didn't make it through this group
        poss_transforms = {t for t, x in eq_under_transform.items() if x}
      # compile equivalences from all groups made under still valid transforms
      # a dict of sets so singletons aren't lost in finding connected components
      equivalents = {i: set() for i in range(self.n_tiles)}
      for group_equivalents in by_group_equivalent_tiles:
        for (tr, tile_i, tile_j) in group_equivalents:
          if tr in poss_transforms:
            equivalents[tile_i].add(tile_j)
      self.tile_matching_transforms = {
        k: v for k, v in self.tile_matching_transforms.items()
        if k in poss_transforms}
      self.tile_transitivity_classes = []
      equivalents = nx.connected_components(nx.from_dict_of_lists(equivalents))
      for c, base_IDs in enumerate(equivalents):
        transitivity_class = []
        for tile in self.tiles:
          if tile.base_ID in base_IDs:
            transitivity_class.append(tile.ID)
            tile.transitivity_class = c
        self.tile_transitivity_classes.append(transitivity_class)
    else:
      # transitivity classes are just the individual tiles
      self.tile_transitivity_classes = []
      for i, tile in enumerate(self.tiles):
        tile.transitivity_class = tile.base_ID
        if i < self.n_tiles:
          self.tile_transitivity_classes.append([tile.ID])
        else:
          self.tile_transitivity_classes[tile.base_ID].append(tile.ID)


  def get_potential_symmetries(
      self,
      ignore_tile_id_labels:bool = True,
    ) -> dict[int, Transform]:
    """Assemble potential symmetries from symmetries of prototile and tiles.

    Also remove any duplicates that result. The result is assigned to the
    tile_matching_transforms attribute.

    TODO: consider retaining the Symmetry objects as these carry additional
    information that might facilitate labelling under a limited number of the
    symmetries not all of them.

    Returns:
      dict[int, tuple[float]]: dictionary of the symmetries (transforms
        actually) in shapely affine transform 6-tuple format.

    """
    self.tile_matching_transforms = {
      k: Transform("translation", 0, geom.Point(0, 0), v,
                   tiling_utils.get_translation_transform(v[0], v[1]))
      for k, v in enumerate(self.tileable.get_vectors())}
    if ignore_tile_id_labels:
      n_symmetries = len(self.tile_matching_transforms)
      ptile = self.tileable.prototile.loc[0, "geometry"]
      for tr in ShapeMatcher(ptile).get_polygon_matches(ptile):
        if tr.transform_type not in ["identity", "translation"]:
          self.tile_matching_transforms[n_symmetries] = tr
          n_symmetries = n_symmetries + 1
      for tile in self.tiles[:self.n_tiles]:
        for tr in ShapeMatcher(tile.shape).get_polygon_matches(tile.shape):
          if tr.transform_type not in ["identity", "translation"]:
            self.tile_matching_transforms[n_symmetries] = tr
            n_symmetries = n_symmetries + 1
      for tile in self.tiles[:self.n_tiles]:
        sm = ShapeMatcher(tile.shape)
        transforms = [sm.get_polygon_matches(self.tiles[i].shape)
          for i in self.shape_groups[tile.shape_group] if i < self.n_tiles]
        for tr in itertools.chain(*transforms):
          if tr.transform_type not in ["identity", "translation"]:
            self.tile_matching_transforms[n_symmetries] = tr
            n_symmetries = n_symmetries + 1
      self.tile_matching_transforms = self._remove_duplicate_symmetries(
        self.tile_matching_transforms)
    return self.tile_matching_transforms


  def _remove_duplicate_symmetries(
      self,
      transforms:dict[int,Transform],
    ) -> dict[int,Transform]:
    """Filter list of shapely affine transforms to remove duplicates.

    Args:
      transforms (dict[int,Transform]): dictionary of Transforms to filter.

    Returns:
      dict[int,Transform]: the filtered dictionary with duplicates removed.

    """
    uniques = {}
    for k, v in transforms.items():
      already_exists = False
      for u in uniques.values():
        already_exists = np.allclose(
          v.transform, u.transform, atol = 1e-4, rtol = 1e-4)
        if already_exists:
          break
      if not already_exists:
        uniques[k] = v
    return uniques


  def _find_vertex_transitivity_classes(
      self,
      ignore_tile_id_labels:bool = True,
    ) -> None:
    """Find vertex transitivity classes.

    This function checks which vertices align with which others under transforms
    in the tile_matching_transforms attribute. The process need only determine
    the classes for vertices in the core tileable.tiles, then assign those to
    all vertices by matched base_ID.
    """
    if ignore_tile_id_labels:
      equivalent_vertices = defaultdict(set)
      base_vertices = [v for v in
                      self.vertices_in_tiles(self.tiles[:self.n_tiles])
                      if v.is_tiling_vertex]
      for transform in self.tile_matching_transforms.values():
        for v in base_vertices:
          equivalent_vertices[v.ID].add(v.ID)
          match_ID = self._match_geoms_under_transform(
            v, base_vertices, transform.transform)
          if match_ID != -1:
            equivalent_vertices[v.ID].add(match_ID)
      equivalent_vertices = self._get_exclusive_supersets(
        [tuple(sorted(s)) for s in equivalent_vertices.values()])
      self.vertex_transitivity_classes = defaultdict(list)
      for c, vclass in enumerate(equivalent_vertices):
        for v in self.points.values():
          if v.base_ID in vclass:
            v.transitivity_class = c
            self.vertex_transitivity_classes[c].append(v.ID)
      self.vertex_transitivity_classes = list(
        self.vertex_transitivity_classes.values())
      # label vertices based on their transitivity class
      for v in self.points.values():
        if v.is_tiling_vertex:
          v.label = LABELS[v.transitivity_class]
    else:
      self.vertex_transitivity_classes = defaultdict(list)
      for v in self.points.values():
        if v.is_tiling_vertex:
          self.vertex_transitivity_classes[v.base_ID].append(v.ID)
          v.transitivity_class = v.base_ID
      self.vertex_transitivity_classes = list(
        self.vertex_transitivity_classes.values())
      for v in self.points.values():
        if v.is_tiling_vertex:
          v.label = LABELS[v.transitivity_class]


  def _find_edge_transitivity_classes(
      self,
      ignore_tile_id_labels:bool = True,
    ) -> None:
    """Find edge transitivity classes.

    This function works by checking which edges align with which others under
    transforms in the tile_matching_transforms attribute. The process need only
    determine the classes for edges in the core tileable.tiles, then assign
    those to all edges by matched base_ID.

    TODO: Note that this code is identical to the vertex transitivity code
    so it might make sense to merge.
    """
    if ignore_tile_id_labels:
      equivalent_edges = defaultdict(set)
      base_edges = self.edges_in_tiles(self.tiles[:self.n_tiles])
      for transform in self.tile_matching_transforms.values():
        for e in base_edges:
          equivalent_edges[e.ID].add(e.ID)
          match_id = self._match_geoms_under_transform(
            e, base_edges, transform.transform)
          if match_id != -1:
            equivalent_edges[e.ID].add(match_id)
      equivalent_edges = self._get_exclusive_supersets(
        [tuple(sorted(s)) for s in equivalent_edges.values()])
      self.edge_transitivity_classes = defaultdict(list)
      for c, eclass in enumerate(equivalent_edges):
        for e in self.edges.values():
          if e.base_ID in eclass:
            e.transitivity_class = c
            self.edge_transitivity_classes[c].append(e.ID)
      self.edge_transitivity_classes = list(
        self.edge_transitivity_classes.values())
      # label edges based on their transitivity class
      for e in self.edges.values():
        e.label = labels[e.transitivity_class]
    else:
      self.edge_transitivity_classes = defaultdict(list)
      for e in self.edges.values():
        self.edge_transitivity_classes[e.base_ID].append(e.ID)
      self.edge_transitivity_classes = list(
        self.edge_transitivity_classes.values())
      for i, eclass in enumerate(self.edge_transitivity_classes):
        for e in eclass:
          self.edges[e].transitivity_class = i
          self.edges[e].label = labels[i]


  def _match_geoms_under_transform(
      self,
      geom1:Tile|Vertex|Edge,
      geoms2:list[Tile|Vertex|Edge],
      transform:tuple[float,...],
    ) -> int|tuple[int]:
    """Determine if a geometry maps onto any in a patch under supplied symmetry.

    Args:
      geom1 (Tile|Vertex|Edge): element whose geometry we want to match.
      geoms2 (list[Tile|Vertex|Edge]): set of elements among which a
        match is sought.
      transform (tuple[float]): shapely affine transform 6-tuple to apply.

    Returns:
      int|tuple[int,int]: ID of the element in patch that matches the geom1
        element under the transform if one exists, otherwise returns -1. For
        edges note that the ID is a tuple.

    """
    match_id = -1
    for geom2 in geoms2:
      if isinstance(geom1, Tile):
        # an area of intersection based test
        match = self.polygon_matches(
          affine.affine_transform(geom1.shape, transform), geom2.shape)
      elif isinstance(geom1, Vertex):
        # distance test
        match = affine.affine_transform(geom1.point, transform).distance(
          geom2.point) <= 10 * tiling_utils.RESOLUTION
      else: # must be an Edge
        # since edges _should not_ intersect this test should work in
        # lieu of a more complete point by point comparison
        c1 = geom1.get_geometry().centroid
        c2 = geom2.get_geometry().centroid
        match = affine.affine_transform(c1, transform) \
          .distance(c2) <= 10 *tiling_utils.RESOLUTION
      if match:
        return geom2.base_ID
    return match_id


  def _get_exclusive_supersets(
      self, sets:list[Iterable]) -> list[Iterable]:
    """Return sets of elements not found in the same set among those supplied.

    The supplied sets share elements, i.e., they are non-exclusives sets. The
    returned sets are exclusive: each element will only appear in one of the
    sets in the returned list. This is accomplished using networkx's
    connected components applied to a graph where each intersection between two
    sets is an edge.

    Args:
        sets (list[Iterable]): list of lists of possibly overlapping sets.

    Returns:
      list[Iterable]: list of lists that include all the original
        elements without overlaps.

    """
    overlaps = []
    for i, si in enumerate(sets):
      s1 = set(si)
      for j, sj in enumerate(sets):
        s2 = set(sj)
        if len(s1 & s2) > 0:
          overlaps.append((i, j))
    G = nx.from_edgelist(overlaps)
    result = []
    for component in nx.connected_components(G):
      s = set()
      for i in component:
        s = s.union(sets[i])
      result.append(tuple(s))
    return result


  def vertices_in_tiles(self, tiles:list[Tile]) -> list[Vertex]:
    """Get vertices incident on tiles in supplied list.

    Args:
      tiles (list[Tile]): tiles whose vertices are required.

    Returns:
      list[Vertex]: the required vertices.

    """
    vs = set()
    for tile in tiles:
      vs = vs.union(tile.get_corner_IDs())
    return [self.points[v] for v in vs]


  def edges_in_tiles(self, tiles:list[Tile]) -> list[Edge]:
    """Get edges that are part of the boundary of tiles in supplied list.

    Args:
      tiles (list[Tile]): tiles whose edges are required.

    Returns:
      list[Edge]: the required edges.

    """
    es = set()
    for tile in tiles:
      es = es.union(tile.get_edge_IDs())
    return [self.edges[e] for e in es]


  def generate_dual(self) -> list[geom.Polygon]:
    """Create the dual tiiing for the tiling of this Topology.

    TODO: make this a viable replacement for the existing dual tiling
    generation.

    TODO: also need to ensure that this finds a set of dual tiles that exhaust
    the plane...

    Returns:
      list[geom.Polygon]: a list of polygon objects.

    """
    for v in self.points.values():
      v.clockwise_order_incident_tiles()
    self.dual_tiles = {}
    base_id_sets = defaultdict(list)
    for v in self.points.values():
      base_id_sets[v.base_ID].append(v.ID)
    minimal_set = [self.points[min(s)] for s in base_id_sets.values()]
    for v in minimal_set:
    # for v in self.points.values():
      if v.is_interior() and len(v.tiles) > 2:
        self.dual_tiles[v.ID] = \
          geom.Polygon([t.centre for t in v.tiles])


  def get_dual_tiles(self) -> gpd.GeoDataFrame:
    """Return dual tiles as GeoDataFrame."""
    n = len(self.dual_tiles)
    return gpd.GeoDataFrame(
      data = {"tile_id": list(self.tileable.tiles.tile_id)[:n]},
      geometry = gpd.GeoSeries(self.dual_tiles.values()),
      crs = self.tileable.crs)


  def add_vertex(self, pt:geom.Point) -> Vertex:
    """Add and return Vertex at the specified point location.

    No attempt is made to ensure Vertex IDs are an unbroken sequence: a new ID
    is generated one greater than the existing highest ID. IDs will usually be
    an unbroken sequence up to removals when geometry transformations are
    applied.

    Args:
      pt (geom.Point): point location of the Vertex.

    Returns:
      Vertex: the added Vertex object.

    """
    n = 0 if len(self.points) == 0 else max(self.points.keys()) + 1
    v = Vertex(pt, n)
    self.points[n] = v
    return v


  def add_edge(self, vs:list[Vertex]) -> Edge:
    """Create an Edge from the suppled vertices and return it.

    The new Edge is added to the edges dictionary. Edges are self indexing by
    the IDs of their end Vertices.

    Args:
      vs (list[Vertex]): list of Vertices in the Edge to be created.

    Returns:
        Edge: the added Edge.

    """
    e = Edge(vs)
    self.edges[e.ID] = e
    return e


  def polygon_matches(
      self,
      geom1:geom.Polygon,
      geom2:geom.Polygon,
    ) -> bool:
    """Test if supplied polygons match geometrically.

    Tests for equality of area, and equality of their area of overlap to their
    shared area, i.e. Area1 == Area2 == (Area 1 intersection 2).

    Args:
      geom1 (geom.Polygon): first polygon.
      geom2 (geom.Polygon): second polygon.

    Returns:
      bool: True if the polygons are the same, False otherwise.

    """
    a, b = geom1.area, geom2.area
    return bool(
      np.isclose(a, b,
                 rtol = tiling_utils.RESOLUTION * 100,
                 atol = tiling_utils.RESOLUTION * 100) and
      np.isclose(a, geom1.intersection(geom2).area,
                 rtol = tiling_utils.RESOLUTION * 100,
                 atol = tiling_utils.RESOLUTION * 100))


  def _get_tile_geoms(self) -> gpd.GeoDataFrame:
    """Return GeoDataFrame of the tiles.

    Returns:
      gpd.GeoDataFrame: the tiles as a GeoDataFrame.

    """
    return gpd.GeoDataFrame(
      data = {"transitivity_class": [t.transitivity_class for t in self.tiles]},
      geometry = gpd.GeoSeries([t.shape for t in self.tiles]))


  def _get_tile_centre_geoms(self) -> gpd.GeoSeries:
    """Return the centre of tiles as a GeoDataFrame.

    Returns:
      gpd.GeoDataFrame: the tiles centres.

    """
    return gpd.GeoSeries([t.centre for t in self.tiles])


  def _get_point_geoms(self) -> gpd.GeoSeries:
    """Return tiling vertices and tile corners as a GeoDataFrame of Points.

    Returns:
      gpd.GeoDataFrame: corners and vertices of the tiling.

    """
    return gpd.GeoSeries([v.point for v in self.points.values()])


  def _get_vertex_geoms(self) -> gpd.GeoSeries:
    """Return tiling vertices (not corners) as a GeoDataFrame of Points.

    Returns:
      gpd.GeoDataFrame: vertices of the tiling.

    """
    return gpd.GeoSeries([v.point for v in self.points.values()
                          if v.is_tiling_vertex])


  def _get_edge_geoms(
      self,
      offset:float = 0.0,
    ) -> gpd.GeoSeries:
    """Return tiling edges a GeoSeries of LineStrings optionally offset.

    Offsetting the edges allows their direction to be seen when visualised.

    Returns:
      gpd.GeoDataFrame: edges of the tiling.

    """
    return gpd.GeoSeries([e.get_topology().parallel_offset(offset)
                          for e in self.edges.values()])


  def plot(
      self,
      show_original_tiles: bool = True,
      show_tile_centres: bool = False,
      show_vertex_labels: bool = True,
      show_vertex_ids: bool = False,
      show_edges: bool = True,
      offset_edges: bool = True,
      show_edge_labels:bool = False,
      show_dual_tiles: bool = False,
    ) -> plt.Axes:
    """Delegate plotting of requested elements and return plt.Axes.

    Args:
      show_original_tiles (bool, optional): if True show the tiles. Defaults to
        True.
      show_tile_centres (bool, optional): if True show tile centres with upper
        case alphabetical labels. Defaults to False.
      show_vertex_labels (bool, optional): if True show tiling vertices labelled
        to show their equivalence classes. Defaults to True.
      show_vertex_ids (bool, optional): if True show vertex IDs (i.e., sequence
        numbers) which is useful for debugging. Defaults to False.
      show_edges (bool, optional): if True show tiling edges (not tile sides).
        Defaults to True.
      offset_edges (bool, optional): if True offset edges a little from and
        parallel to their geometric position. Defaults to True.
      show_edge_labels (bool, optional): if true show lower case alphabetical
        labels identifying edge equivalence classes. Defaults to False.
      show_dual_tiles (bool, optional): if True show a candidate set of dual
        tiles as an overlay. Defaults to False.

    Returns:
      plt.Axes: a plot of the Topology as requested.

    """
    fig = plt.figure(figsize = (10, 10))
    ax = fig.add_axes(111)
    extent = gpd.GeoSeries([t.shape for t in self.tiles]).total_bounds
    dist = max([extent[2] - extent[0], extent[3] - extent[1]]) / 100
    if show_original_tiles:
      self._plot_tiles(ax)
    if show_tile_centres:
      self._plot_tile_centres(ax)
    if show_vertex_labels:
      self._plot_vertex_labels(ax, show_vertex_ids)
    if show_edge_labels or show_edges:
      self._plot_edges(ax, show_edges, show_edge_labels, dist, offset_edges)
    if show_dual_tiles:
      self._plot_dual_tiles(ax, dist)
    plt.axis("off")
    return ax


  def _plot_tiles(self, ax:plt.Axes) -> plt.Axes:
    """Plot Topology's tiles on supplied Axes.

    Tiles are coloured by equivalence class.

    Args:
      ax (plt.Axes): Axes on which to plot.

    Returns:
      plt.Axes: the Axes.

    """
    self._get_tile_geoms().plot(column = "transitivity_class",
      ax = ax, ec = "#444444", lw = 0.5, alpha = 0.25, cmap = "Greys")
    return ax


  def _plot_tile_centres(self, ax:plt.Axes) -> plt.Axes:
    """Print tile transitivity class at each tile centroid.

    Args:
      ax (plt.Axes): Axes on which to plot.

    Returns:
      plt.Axes: the Axes.

    """
    for _, tile in enumerate(self.tiles):
      ax.annotate(tile.transitivity_class, xy = (tile.centre.x, tile.centre.y),
                  ha = "center", va = "center")
    return ax


  def _plot_vertex_labels(
      self,
      ax:plt.Axes,
      show_vertex_ids:bool = False,
    ) -> plt.Axes:
    """Plot either the Vertex transitivity class label or its sequential ID.

    Args:
      ax (plt.Axes): Axes on which to plot.
      show_vertex_ids (bool, optional): If True plots the ID, else plots the
        transitivity class. Defaults to False.

    Returns:
      plt.Axes: the Axes.

    """
    for v in self.points.values():
      ax.annotate(v.ID if show_vertex_ids else v.label,
                  xy = (v.point.x, v.point.y), color = "k",
                  ha = "center", va = "center")
    return ax


  def _plot_edges(
      self,
      ax:plt.Axes,
      show_edges:bool = False,
      show_edge_labels:bool = False,
      dist:float = 0.0,
      offset_edges:bool = True,
    ) -> plt.Axes:
    """Plot edges, including an offset if specified and labels if requested.

    Can also be called to only plot the labels.

    Args:
      ax (plt.Axes): Axes on which to plot.
      show_edges (bool, optional): if True includes the edges as. a dotted blue
        line, optionally offset (for clarity) from the tile boundary. Defaults
        to False.
      show_edge_labels (bool, optional): if True shows an edge label. Defaults
        to False.
      dist (float, optional): a distance by which to offset the dotted line for
        the edge from the tile boundary. Defaults to 0.0.
      offset_edges (bool, optional): if True applies the edge drawing offset,
        if False the edge is drawn as a single line segment between its end
        vertices (and may not align with the sides of tiles. Defaults to True.

    Returns:
      plt.Axes: the Axes.

    """
    if show_edges:
      edges = self._get_edge_geoms(dist if offset_edges else 0)
      edges.plot(ax = ax, color = "dodgerblue", ls = ":")
    else:
      edges = [e.get_geometry() for e in self.edges.values()]
    if show_edge_labels:
      for ls, e in zip(edges, self.edges.values(), strict = True):
        c = ls.centroid
        ax.annotate(e.label, xy = (c.x, c.y), color = "k",
                    ha = "center", va = "center")
    return ax

  def _plot_dual_tiles(self, ax:plt.Axes,
                       dist:float = 0.0) -> plt.Axes:
    gpd.GeoSeries(self.dual_tiles).buffer(
      -dist / 4, join_style = "mitre", cap_style = "square").plot(
        ax = ax, fc = "g", alpha = 0.25)
    return ax


  def plot_tiling_symmetries(
      self,
      **kwargs:float,
    ) -> None:
    """Plot the symmetries of Topology's tiling.

    Most of the work here is delegated to `_plot_tiling_symmetry` which is run
    once per symmetry on a grid of plt.Axes built by this function.

    Args:
      kwargs: passed through to `_plot_tiling_symmetry`

    """
    n = len(self.tile_matching_transforms)
    nc = int(np.ceil(np.sqrt(n)))
    nr = int(np.ceil(n / nc))
    fig = plt.figure(figsize = (12, 12 * nr / nc))
    for i, tr in enumerate(self.tile_matching_transforms.values()):
      ax = fig.add_subplot(nr, nc, i + 1)
      self._plot_tiling_symmetry(tr, ax, **kwargs)


  def _plot_tiling_symmetry(
      self,
      transform:Transform,
      ax:plt.Axes,
      **kwargs:float,
    ) -> None:
    """Plot the supplied Transform on the supplied plt.Axes.

    Args:
        transform (Transform): the Transform to plot.
        ax (plt.Axes): the Axes on which to plot.
        kwargs: passed through to plt.plot functions.

    """
    tiles = gpd.GeoSeries([t.shape for t in self.tiles])
    base_tiles = tiles[:self.n_tiles]
    tiles.plot(ax = ax, fc = "k", alpha = .15, ec = "k", lw = .5)
    base_tiles.plot(ax = ax, fc = "#00000000", ec = "w", lw = 1, zorder = 2)
    transformed_base_tiles = gpd.GeoSeries(
      [transform.apply(g) for g in base_tiles])
    transformed_base_tiles.plot(ax = ax, fc = "k", alpha = .2, lw = 0, ec = "k")
    transform.draw(ax, **kwargs) # delegate to Transform.draw
    plt.axis("off")


  def transform_geometry(
      self,
      new_topology:bool,
      apply_to_tiles:bool,
      selector:str,
      transform_type:str,
      **kwargs:float) -> Topology:
    r"""Get a new Topology by applying specified transformation.

    A transformation specified by `transform_type` and keyword arguments is
    applied to elements in the Topology whose labels match the selector
    parameter. The transform is optionally applied to update tiles and
    optionally requests a new Topology object.

    Implemented in this way so that transformations can be applied one at a time
    without creating an intermediate set of new tiles, which may be invalid and
    fail. So, if you wish to apply (say) 3 transforms and generate a new
    Topology leaving the existing one intact:

        new_topo = old_topo.transform_geometry(True,  False, "a", ...) \
                           .transform_geometry(False, False, "B", ...) \
                           .transform_geometry(False, True,  "C", ...)

    The first transform requests a new Topology, subsequent steps do not, and it
    is only the last step which attempts to create the new tile polygons.

    **kwargs supply named parameters for the requested transformation.

    Args:
      new_topology (bool): if True returns a new Topology object, else returns
        the current Topology modified.
      apply_to_tiles (bool): if True attempts to create new Tiles after the
        transformation has been applied. Usually set to False, unless the last
        transformation in a pipeline, to avoid problems of topologically invalid
        tiles at intermediate steps.
      selector (str): label of elements to which to apply the transformation.
        Note that all letters in the supplied string are checked, so you can
        use e.g. "abc" to apply a transformation to edges labelled "a", "b" or
        "c", or "AB" for vertices labelled "A" or "B".
      transform_type (str): name of the type of transformation requested.
        Currently supported are `zigzag_edge`, `rotate_edge`, `scale_edge`,
        `push_vertex`, and `nudge_vertex`. Keyword arguments for each are
        documented in the corresponding methods.
      kwargs: contains any needed arguments of the requested transform_type.

    Returns:
      Topology: if new_topology is True a new Topology based on this one with
        after transformation, if False this Topology is returned after the
        transformation.

    """
    print("CAUTION: new Topology will probably not be correctly labelled. "
          "To build a correct Topology, extract the tileable attribute and "
          "rebuild Topology from that.")
    # pickle seems to be more reliable than copy.deepcopy here
    topo = (pickle.loads(pickle.dumps(self))   # noqa: S301
            if new_topology
            else self)
    transform_args = topo.get_kwargs(getattr(topo, transform_type), **kwargs)
    match transform_type:
      case "zigzag_edge":
        for e in topo.edges.values():
          if e.label in selector:
            topo.zigzag_edge(e, **transform_args)
      case "rotate_edge":
        for e in topo.edges.values():
          if e.label in selector:
            topo.rotate_edge(e, **transform_args)
      case "scale_edge":
        for e in topo.edges.values():
          if e.label in selector:
            topo.scale_edge(e, **transform_args)
      case "push_vertex":
        pushes = {}
        for v in topo.vertices_in_tiles(topo.tiles[:topo.n_tiles]):
          if v.label in selector:
            pushes[v.base_ID] = topo.push_vertex(v, **transform_args)
        for base_ID, (dx, dy) in pushes.items():
          for v in [v for v in topo.points.values() if v.base_ID == base_ID]:
            v.point = affine.translate(v.point, dx, dy)
      case "nudge_vertex":
         for v in topo.points.values():
          if v.label in selector:
            topo.nudge_vertex(v, **transform_args)
    if apply_to_tiles:
      for t in topo.tiles:
        t.set_corners_from_edges()
    topo.tileable.tiles.geometry = gpd.GeoSeries(
      [topo.tiles[i].shape for i in range(topo.n_tiles)])
    topo.tileable._setup_regularised_prototile()
    return topo


  def get_kwargs(
      self,
      fn:Callable,
      **kwargs:str|float,
    ) -> dict:
    """Filter the supplied kwargs to only contain arguments required by fn.

    Args:
      fn (Callable): the function that is to be inspected.
      **kwargs (str|float): kwargs to be filtered.

    Returns:
      str|float: filtered dictionary of kwargs.

    """
    args = inspect.signature(fn).parameters
    return {k: kwargs.pop(k) for k in dict(kwargs) if k in args}


  def zigzag_edge(
      self,
      edge:Edge,
      start:str = "A",
      n:int = 2,
      h:float = 0.5,
      smoothness:int = 0,
    ) -> None:
    """Apply zigzag transformation to supplied Edge.

    Currently this will only work correctly if h is even.

    TODO: make it possible for odd numbers of 'peaks' to work (this may require
    allowing bidirectional Edges, i.e. storing Edges in both directions so that
    all Tile edges are drawn CW). The `start` parameter is a temporary hack for
    this.

    Args:
      edge (Edge): Edge to transform
      start (str, optional): label at one end of edge which is used to determine
        the sense of h, enabling C-curves with an odd number n of zigs and zags
        to be applied. Defaults to 'A'.
      n (int, optional): number of zigs and zags in the edge. Defaults to 2.
      h (float, optional): width of the zig zags relative to edge length.
        Defaults to 0.5.
      smoothness (int, optional): spline smoothness. 0 gives a zig zag proper,
        higher values will produce a sinusoid. Defaults to 0.

    """
    v0, v1 = edge.vertices[0], edge.vertices[1]
    if n % 2 == 1 and v0.label != start:
      h = -h
    ls = self.zigzag_between_points(v0.point, v1.point, n, h, smoothness)
    # remove current corners
    self.points = {k: v for k, v in self.points.items()
                   if k not in edge.get_corner_IDs()[1:-1]}
    # add the new ones
    new_corners = [self.add_vertex(geom.Point(xy)) for xy in ls.coords]
    edge.corners = edge.vertices[:1] + new_corners + edge.vertices[-1:]
    if edge.right_tile is not None:
      edge.right_tile.set_corners_from_edges(False)
    if edge.left_tile is not None:
      edge.left_tile.set_corners_from_edges(False)


  def zigzag_between_points(
      self,
      p0:geom.Point,
      p1:geom.Point,
      n:int,
      h:float = 1.0,
      smoothness:int = 0,
    ) -> geom.LineString:
    """Return a zig zag line optionally smoothed as a spline between two points.

    Args:
      p0 (geom.Point): start point.
      p1 (geom.Point): end point.
      n (int): number of zig zags.
      h (float, optional): amplitude of zig zags relative to distance between
        points. Defaults to 1.0.
      smoothness (int, optional): number of spline smoothed points to add. If
        set to 0 a straight line zig zag is produced. Defaults to 0.

    Returns:
      geom.LineString: the resulting zig zag line.

    """
    r = p0.distance(p1)
    # make a sinusoidal template
    x = np.linspace(0, n * np.pi, n * 2 + 1, endpoint = True)
    y = [np.sin(x) for x in x]
    spline = interpolate.InterpolatedUnivariateSpline(x, y, k = 2)

    spline_steps = (n + smoothness) * 2 + 1
    xs = np.linspace(0, n * np.pi, spline_steps, endpoint = True)
    ys = spline(xs)

    sfx = 1 / max(x) * r
    sfy = h * r / 2
    theta = np.arctan2(p1.y - p0.y, p1.x - p0.x)

    ls = geom.LineString(
      [geom.Point(x, y) for x, y in zip(xs, ys, strict = True)])
    ls = affine.translate(ls, 0, -(ls.bounds[1] + ls.bounds[3]) / 2)
    ls = affine.scale(ls, xfact = sfx, yfact = sfy, origin = (0, 0))
    ls = affine.rotate(ls, theta, (0, 0), use_radians = True)
    x0, y0 = next(iter(ls.coords))
    return affine.translate(ls, p0.x - x0, p0.y - y0)


  def rotate_edge(
      self,
      edge:Edge,
      centre:str = "",
      angle:float = 0,
    ) -> None:
    """Rotate edge.

    Args:
      edge (Edge): the edge to rotate.
      centre (str): centre of rotation which should be label of one of its
        end point vertices or "" when rotation will be about the Edge centroid.
        Defaults to "".
      angle (float): angle of rotation.

    """
    v0, v1 = edge.vertices
    ls = geom.LineString([v0.point, v1.point])
    if v0.label == centre:
      c = v0.point
    elif v1.label == centre:
      c = v1.point
    else:
      c = ls.centroid
    ls = affine.rotate(ls, angle, origin = c)
    v0.point, v1.point = [geom.Point(c) for c in ls.coords]


  def scale_edge(
      self,
      edge:Edge,
      sf:float = 1.0,
    ) -> None:
    """Scale edge.

    Args:
      edge (Edge): the edge to scale.
      sf (float): amount by which to scale the edge.

    """
    v0, v1 = edge.vertices
    ls = geom.LineString([v0.point, v1.point])
    ls = affine.scale(ls, xfact = sf, yfact = sf, origin = ls.centroid)
    v0.point, v1.point = [geom.Point(c) for c in ls.coords]


  def push_vertex(
      self,
      vertex:Vertex,
      push_d:float,
    ) -> tuple[float]:
    """Return displacement vector to push a vertex based on incident edges.

    Args:
      vertex (Vertex): the vertex to push.
      push_d (float): the distance to push it.

    """
    neighbours = [self.points[v] for v in vertex.neighbours]
    dists = [vertex.point.distance(v.point) for v in neighbours]
    x, y = vertex.point.x, vertex.point.y
    unit_vectors = [((x - v.point.x) / d, (y - v.point.y) / d)
                    for v, d in zip(neighbours, dists, strict = True)]
    return  (push_d * sum([xy[0] for xy in unit_vectors]),
             push_d * sum([xy[1] for xy in unit_vectors]))


  def nudge_vertex(
      self,
      vertex:Vertex,
      dx:float,
      dy:float,
    ) -> None:
    """Nudge vertex by specified displacement.

    Args:
      vertex (Vertex): the vertext to nudge.
      dx (float): x displacement.
      dy (float): y displacement.

    """
    vertex.point = affine.translate(vertex.point, dx, dy)


class Tile:
  """Class to represent essential features of polygons in a tiling."""

  ID: int
  """integer ID number which indexes the Tile in the containing Topology tiles
  list."""
  base_ID: int
  """ID of corresponding Tile in the base tileable unit"""
  corners: list[Vertex]
  """list of Vertex objects. This includes all corners of the original polygon
  and any tiling vertices induced by (for example) a the corner of an adjacent
  tile lying halfway along an edge of the original polygon on which this tile
  is based. Vertex objects are stored in strictly clockwise sequence."""
  edges: list[Edge]
  """list of Edge objects that together compose the tile boundary."""
  edges_CW: list[bool]
  """list of Edge direction. Edges are stored only once in a Topology so some
  edges are in clockwise order and others  are in counter-clockwise order.
  These boolean flags are True if the corresponding Edge is clockwise, False if
  counter-clockwise."""
  label: str
  """tile_id label from the tileable source"""
  vertex_labels: list[str]
  """list of (upper case) letter labels of the tile corners (i.e. all corners,
  not only tiling vertices)."""
  edge_labels: list[str]
  """list of (lower case) letter labels of the tile edges (tiling edges, not
  tile sides)."""
  shape: geom.Polygon = None
  """the tile geometry (which may include some redundant points along sides
  where neighbouring tiles induce a tiling vertex). So for example a rectangle
  might have additional points along its sides:

        +---+-------+
        |   |   2   |
        | 1 A---B---E---+
        |   |   |   4   |
        +---C 3 D-------+
            |   |
            +---+

  In the above Tile 1 has additional point A, 2 has B and 3 has C and D induced
  by the corners of neighbouring tiles."""
  centre: geom.Point = None
  """a point centre for the Tile (determined by weavingspace.tiling_utils.
  incentre)."""
  shape_group: int = None
  """the tile shape group of this tile in its containing Topology."""
  transitivity_class: int = None
  """the tile transitivity class of this tile its containing Topology"""

  def __init__(
      self,
      ID:int,
    ) -> None:
    """Class constructor.

    Args:
      ID (int): sequence number ID of this vertex in the Topology.

    """
    self.ID = ID
    self.corners = []
    self.edges = []
    self.edges_CW = []
    self.vertex_labels = []
    self.edge_labels = []


  def __str__(self) -> str:
    """Return string representation of the Tile.

    Returns:
      str: string including Tile ID, list of corner vertex IDs and list of
        edge IDs.

    """
    return (f"Tile {self.ID} Corners: {self.get_corner_IDs()} "
            f"Edges: {self.get_edge_IDs()}")


  def __repr__(self) -> str:
    return str(self)


  def get_corner_IDs(self) -> list[int]:
    """Return list of corner IDs (not Vertex objects).

    Returns:
      list[int]: list of integer IDs of tile corners.

    """
    return [c.ID for c in self.corners]


  def get_edge_IDs(self) -> list[tuple[int,int]]:
    """Return list of edge IDs (not Edge objects).

    Returns:
      list[tuple[int]]: list of 2-element tuples of the start and end Vertex IDs
        of each edge.

    """
    return [e.ID for e in self.edges]


  def set_shape_from_corners(self) -> None:
    """Set the shape attribute based on corners, and associated tile centre."""
    self.shape = geom.Polygon([c.point for c in self.corners])
    self.centre = tiling_utils.get_incentre(
      tiling_utils.get_clean_polygon(self.shape))


  def set_corners_from_edges(
      self,
      update_shape:bool = True,
    ) -> None:
    """Set corners attribute from the edges attribute.

    Typically called after modification of topology edges. Optionally the shape
    attribute is NOT updated, which may save time when multiple changes to the
    edges of a tile are in process (i.e., only update the shape after all
    changes are complete).

    Args:
      update_shape (bool, optional): if True the shape attribute will be
        updated, otherwise not. Defaults to True.

    """
    self.corners = []
    for e, cw in zip(self.edges, self.edges_CW, strict = True):
      if cw: # clockwise to extend by all but the first corner
        self.corners.extend(e.corners[:-1])
      else: # counter-clockwise so extend in reverse
        self.corners.extend(e.corners[1:][::-1])
    if update_shape:
      self.set_shape_from_corners()


  def set_edge_directions(self) -> None:
    """Set up edges_CW attribute by inspection of the edges list.

    It is (frankly!) hard to keep track of the correct sequence of CW/CCW order
    of edges as new ones are created or old ones merged. This method inspects
    the 'tail-head' relations between consecutive edges to set these flags
    correctly.

    The test is simply to check if the 'tail' Vertex ID in each edge appears
    in the ID tuple of the following edge, i.e. if successive edge
    IDs are (0, 1) (2, 1) or (0, 1) (1, 2), then edge (0, 1) is in clockwise
    direction, but if we have (0, 1) (2, 3) then it is not.
    """
    edge_IDs = self.get_edge_IDs()
    self.edges_CW = [e1[-1] in e2 for e1, e2 in
                     zip(edge_IDs, edge_IDs[1:] + edge_IDs[:1], strict = True)]


  def insert_vertex_at(
      self,
      v:Vertex,
      i:int,
      update_shape:bool = False,
    ) -> tuple[tuple[int,int],tuple[tuple[int,int],...]]:
    """Insert the Vertex into tile at index position i.

    Both corners and edges attributes are updated, and the old edge IDs for
    removal and the new edge itself are returned to the calling context (the
    containing Topology) for update of its edges collection. Optionally update
    the shape attribute.

    This is NOT a generic vertex insertion method: it is only for use during
    Topology initialisation, and does not guarantee correct maintenance of
    all tile, edge and vertex relations in the general case---at any rate it
    has not been tested for this!

    Args:
      v (Vertex): the Vertex to insert.
      i (int): index position in current corners after which to insert
        supplied Vertex.
      update_shape (bool, optional): if True shape attribute is updated.
        Defaults to False.

    Returns:
      tuple: the (tuple) ID of the old edge which should be deleted, and
        the new Edges arising from insertion of this Vertex.

    """
    self.corners = self.corners[:i] + [v] + self.corners[i:]
    old_edge = self.edges[i - 1]
    # store current ID of the affected edge for return to calling context
    old_edge_ID = old_edge.ID
    new_edges = old_edge.insert_vertex(v, self.corners[i - 1])
    self.edges = self.edges[:(i-1)] + new_edges + self.edges[i:]
    self.set_edge_directions()
    if update_shape:
      self.set_shape_from_corners()
    return old_edge_ID, new_edges


  def merge_edges_at_vertex(
      self,
      v:Vertex,
    ) -> tuple:
    """Merge edges that meet at the supplied Vertex.

    It is assumed that only two tiles are impacted this one, and its neighbour
    across the Edge on which v lies. Both are updated. For this reason the work
    is delegated to `get_updated_edges_from_merge` which is run on both affected
    tiles, but only determines the edges to remove and the new edge to be added
    once. See that method for details.

    Args:
      v (Vertex): Vertex at which to merge Edges. This should currently be an
        end

    Returns:
      tuple: 2 item list of the edge IDs to be removed and a new Edge object to
        be added by the calling context (i.e. the containing Topology).

    """
    to_remove, new_edge = self.get_updated_edges_from_merge(v)
    if len(v.tiles) > 1:
      v.tiles[1].get_updated_edges_from_merge(v, new_edge)
    return to_remove, new_edge


  def get_updated_edges_from_merge(
      self,
      v:Vertex,
      new_edge:Edge = None,
    ) -> None|tuple[tuple[tuple[int,int], tuple[int,int]], Edge]:
    """Update edges and edges_CW attributes based on insertion of Vertex.

    If new_edge is supplied then the neighbour tile at v has already created
    the needed new Edge and this Edge is the one that will be 'slotted in' at
    the appropriate spot in the edges list.

    The edges_CW is also updated to maintain correct directions of the edges.
    The corners attribute is unaffected by these changes.

    Args:
      v (Vertex): existing Vertex at which to carry out the merge.
      new_edge (Edge, optional): if another Tile has already carried out this
        merge this should be the resulting new Edge for insertion into this
        Tile. Defaults to None (when the new Edge will be constructed).

    Returns:
      None|tuple: either None (if a new edge was supplied) or a tuple
        of the two edge IDs to be removed and the new edge added for return to
        the calling context (i.e. the containing Topology).

    """
    # get the two edge list index positions in which Vertex v is found
    i, j = self.get_edge_IDs_including_vertex(v)
    if new_edge is None: # then we must make a new one
      # also record existing edge IDs to be removed
      to_remove = [self.edges[i].ID, self.edges[j].ID]
      new_edge = self.get_merged_edge(i, j)
      return_edge_updates = True
    else:
      return_edge_updates = False
    if abs(i - j) != 1:
      # edge indices 'wrap' around from end of edge list to start so drop
      # first and last current edges and stick new one on at the end
      self.edges = self.edges[1:-1] + [new_edge]
    else:
      # insert new edge into list in place of the two old ones
      self.edges = self.edges[:i] + [new_edge] + self.edges[j+1:]
    # update the edge directions
    self.set_edge_directions()
    if return_edge_updates:
      return to_remove, new_edge
    return None


  def get_edge_IDs_including_vertex(
      self,
      v:Vertex,
    ) -> tuple[int]:
    """Get the (two) index positions of the edges that include supplied Vertex.

    Args:
        v (Vertex): Vertex of interest.

    Returns:
      tuple[int]: two index positions of Edges in edges list that contain v.

    """
    return (i for i, e in enumerate(self.edges) if v.ID in e.ID)


  def get_merged_edge(self, i:int, j:int) -> Edge:
    """Return edge made by merging existing edges at i and j in the edges list.

    For example, if the current list of edge IDs was

        (0 1 2) (4 2) (4 5) (5 0)

    and the merge requested is 0 and 1, the resulting new edge is constructed
    from vertices (0 1 2 4).

    Returns:
      Edge: the requested new Edge.

    """
    # if i and j are not consecutive, then j is predecessor edge
    if abs(i - j) != 1:
      i, j = j, i
    # get edges and their directions
    ei, ej = self.edges[i], self.edges[j]
    CWi, CWj = self.edges_CW[i], self.edges_CW[j]
    # DON'T MESS WITH THIS!!!
    # for predecessors (the head) we want everything including the Vertex
    # where the merge is occurring; for successors (the tail) we want all but
    # the first Vertex (which is the one where the merge is occurring). In both
    # cases contingent on whether existing Edges are CW or CCW we may need to
    # flip the Vertex sequence to ensure that the merge Vertex is in the middle
    # of the new edge that will be created
    head = ei.corners if CWi else ei.corners[::-1]
    tail = ej.corners[1:] if CWj else ej.corners[::-1][1:]
    v_sequence = (head if CWi else head[::-1]) + (tail if CWj else tail[::-1])
    return Edge(v_sequence)


  def offset_corners(
      self,
      offset:int,
    ) -> None:
    """Shift shape, corners, edges, and edges_CW by an offset amount.

    This is used to align tiles that are similar, which is required for correct
    transfer of 'base' tile labelling on to 'radius 1' tiles during Topology
    construction.

    Args:
      offset (int): the number of positions to shift the lists.

    """
    if offset is not None or offset != 0:
      self.corners = self.corners[offset:] + self.corners[:offset]
      self.shape = geom.Polygon([c.point for c in self.corners])
      self.edges = self.edges[offset:] + self.edges[:offset]
      self.edges_CW = self.edges_CW[offset:] + self.edges_CW[:offset]


  def get_edge_label(self, e:Edge) -> str:
    """Return edge label of specified edge.

    Args:
        e (Edge): Edge whose label is required.

    Returns:
        str: requested edge label.

    """
    return self.edge_labels[self.get_edge_IDs().index(e.ID)]


  def get_corner_label(self, v:Vertex) -> str:
    """Return corner label of specified corner.

    Args:
        v (Vertex): corner whose label is required.

    Returns:
        str: requested corner label.

    """
    return self.edge_labels[self.get_corner_IDs().index(v.ID)]


  def angle_at(self, v:Vertex) -> float:
    """Return interior angle at the specified corner (in degrees).

    Args:
        v (Vertex): corner where angle is requested.

    Returns:
        float: angle at corner v in degrees.

    """
    i = self.corners.index(v)
    n = len(self.corners)
    return tiling_utils.get_inner_angle(self.corners[i-1].point,
                                        self.corners[i].point,
                                        self.corners[(i + 1) % n].point)


class Vertex:
  """Class to store attributes of a vertex in a tiling."""

  point: geom.Point
  """point (geom.Point): point location of the vertex."""
  ID: int
  """integer (mostly but not necessarily in sequence) of vertex keyed into the
  points dictionary of the containing Topology."""
  tiles: list[Tile]
  """list of Tiles incident on this vertex."""
  neighbours: list[int]
  """list of the immediately adjacent other corner IDs. Only required to
  determine if a point is a tiling vertex (when it will have) three or more
  neighbours, so only IDs are stored."""
  base_ID: int = 1_000_000
  """ID of corresponding Vertex in the tileable base_unit"""
  transitivity_class: int = None
  """transitivity class of the vertex under symmetries of the tiling"""
  label: str = ""
  """the (upper case letter) label of the vertex under the symmetries of the
  tiling."""
  is_tiling_vertex: bool = True
  """is_tiling_vertex (bool): True if this is a tiling vertex, rather than a
  tile corner. E.g., A below is a corner, not a tiling vertex. B is a tiling
  vertex:

      +-------+
      | 1     |
      |   A---B---+
      |   | 2     |
      +---C   +---+
          |   |
          +---+"""


  def __init__(
      self,
      point:geom.Point,
      ID:int,
    ) -> None:
    """Class constructor.

    Args:
      point (geom.Point): point location of the vertex.
      ID (int): a unique integer ID (which will be its key in the containing
        Topology points dictionary).

    """
    self.point = point
    self.ID = ID
    self.base_ID = self.ID
    self.tiles = []
    self.neighbours = []


  def __str__(self) -> str:
    """Return string representation of Vertex.

    Returns:
        str: string including ID, point and list of incident Tiles.

    """
    return f"Vertex {self.ID} at {self.point} Tiles: {self.get_tile_IDs()}"


  def __repr__(self) -> str:
    return str(self)


  def get_tile_IDs(self) -> list[int]:
    """Return list of Tile IDs (not the Tiles themselves).

    Returns:
        list[int]: list of Tile IDs

    """
    return [t.ID for t in self.tiles]


  def add_tile(self, tile:Tile) -> None:
    """Add supplied Tile to the tiles list if it is not already present.

    Args:
        tile (Tile): Tile to add.

    """
    if tile not in self.tiles:
      self.tiles.append(tile)


  def add_neighbour(
      self,
      vertex_id:int,
    ) -> None:
    """Add supplied ID to the neighbours list if it is not already present.

    Args:
      vertex_id (int): ID to add to the neighbours list.

    """
    if vertex_id not in self.neighbours:
      self.neighbours.append(vertex_id)


  def clockwise_order_incident_tiles(self) -> None:
    """Reorder tiles list clockwise (this is for dual tiling construction)."""
    cw_order = self._order_of_pts_cw_around_centre(
      [t.centre for t in self.tiles], self.point)
    self.tiles = [self.tiles[i] for i in cw_order]


  def is_interior(self) -> bool:
    """Test if vertex is completely enclosed by its incident Tiles.

    Based on summing the interior angles of the incident tiles at this vertex.

    Returns:
        bool: True if vertex is completely enclosed by incident Tiles.

    """
    return abs(360 - sum([t.angle_at(self) for t in self.tiles])) \
                     < tiling_utils.RESOLUTION


  def _order_of_pts_cw_around_centre(
      self,
      pts:list[geom.Point],
      centre:geom.Point,
    ) -> list[int]:
    """Return order of points clockwise relative to centre point.

    Args:
      pts (list[geom.Point]): list of points to order.
      centre (geom.Point): centre relative to which CW order is determined.

    Returns:
      list[int]: list of indices of reordered points.

    """
    dx = [p.x - centre.x for p in pts]
    dy = [p.y - centre.y for p in pts]
    angles = [np.arctan2(dy, dx) for dx, dy in zip(dx, dy, strict = True)]
    d = dict(zip(angles, range(len(pts)), strict = True))
    return [i for angle, i in sorted(d.items(), reverse=True)]


class Edge:
  """Class to represent edges in a tiling (not tile sides)."""

  ID: tuple[int]
  """IDs of the vertices at ends of the edge. Used as key in the containing
  Topology's edges dictionary."""
  vertices: list[Vertex]
  """two item list of the end vertices."""
  corners: list[Vertex]
  """list of all the vertices in the edge (including its end vertices). In a
  'normal' edge to edge tiling corners and vertices will be identical."""
  right_tile: Tile = None
  """the tile to the right of the edge traversed from its first to its last
  vertex. Given clockwise winding default, all edges will have a right_tile."""
  left_tile: Tile = None
  """the tile to the left of the edge traversed from its first to its last
  vertex. Exterior edges of the tiles in a Topology will not have a left_tile.
  """
  base_ID: tuple[int] = (1_000_000, 1_000_000)
  """ID of corresponding edge in the base tileable"""
  transitivity_class: int = None
  """transitivity class of the edge under symmetries of the tiling"""
  label: str = ""
  """the (lower case letter) label of the edge under the symmetries of the
  tiling."""

  def __init__(
      self,
      corners:list[Vertex],
    ) -> None:
    """Class constructor.

    Initialises the corners and vertices lists and sets ID to (vertices[0].ID,
    vertices[1].ID). The vertices list is all the corners with is_tiling_vertex
    property True -- Note that during initialisation the default of this
    property is True until after the relations between tiles and vertices have
    been determined.

    Args:
      corners (list[Vertex]): list of all corners along the edge.

    """
    self.corners = corners
    self.vertices = [v for v in self.corners if v.is_tiling_vertex]
    self.ID = tuple(v.ID for v in self.vertices)


  def __str__(self) -> str:
    """Return a string representation of the Edge.

    Returns:
      str: include ID and a list of corner vertex IDs.

    """
    return f"Edge {self.ID} Corners: {[c.ID for c in self.corners]}"


  def __repr__(self) -> str:
    return str(self)


  def get_corner_IDs(self) -> list[int]:
    """Return the IDs of edge corners.

    Returns:
      list[int]: IDs of all corners.

    """
    return [c.ID for c in self.corners]


  def get_vertex_IDs(self) -> list[int]:
    """Return IDs of edge vertices.

    Returns:
      list[int]: list of IDs of the vertices.

    """
    return [v.ID for v in self.vertices]


  def insert_vertex(self, v:Vertex, predecessor:Vertex) -> list[Edge]:
    """Insert vertex after predecessor and return modified edge and new edge.

    If the initial edge was (say) (0 1 2 5) and the predecessor was set to 1
    the returned edges would be (0 1 v) and (v 2 5).
    """
    i = self.corners.index(predecessor)
    new_edge = Edge([v] + self.corners[(i+1):])
    if self.right_tile is not None:
      new_edge.right_tile = self.right_tile
    if self.left_tile is not None:
      new_edge.left_tile = self.left_tile
    self.corners = self.corners[:(i+1)] + [v]
    self.vertices = [self.vertices[0], v]
    self.ID = tuple(v.ID for v in self.vertices)
    return [self, new_edge]


  def get_geometry(
      self,
      forward:bool = True,
    ) -> geom.LineString:
    """Return geom.LineString representing geometry (including corners).

    Args:
      forward (bool, optional): if True the returned LineString starts at
        corners[0], else at corners[-1]. Defaults to True.

    Returns:
      geom.LineString: the required LineString.

    """
    if forward:
      return geom.LineString([v.point for v in self.corners])
    return geom.LineString([v.point for v in self.corners[::-1]])


  def get_topology(
      self,
      forward:bool = True,
    ) -> geom.LineString:
    """Return LineString connecting vertices of this edge.

    Args:
      forward (bool, optional): if True LineString starts at vertices[0],
        else at vertices[1]. Defaults to True.

    Returns:
        geom.LineString: the required LineString.

    """
    if forward:
      return geom.LineString([v.point for v in self.vertices])
    return geom.LineString([v.point for v in self.vertices[::-1]])



"""Classes for working with symmetries."""

from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING

import geopandas as gpd
import numpy as np
import shapely
import shapely.affinity as affine
import shapely.geometry as geom
from matplotlib import pyplot as plt

from weavingspace import tiling_utils

if TYPE_CHECKING:
  from collections.abc import Sequence


@dataclass(slots=True)
class KMP_Matcher:  # noqa: N801
  """Class to find matching subsequences in a sequence."""

  sequence:Sequence
  """Iterable in which subsequences are to be found."""

  def find_matches(
      self,
      pattern:Sequence[tuple[float,...]]) -> list[int]:
    """Find matches in supplied pattern using Knuth-Morris-Pratt algorithm.

    See:
    https://en.wikipedia.org/wiki/Knuth-Morris-Pratt_algorithm which provides
    detailed pseudo-code on which this code is directly based. See also:

    Knuth DE, JH Morris Jr, and VR Pratt. 1977. Fast pattern  matching in
    strings. SIAM Journal on Computing 6(2): 323-350. doi: 10.1137/0206024.

    This implementation expects sequences of tuples of floats, although in
    principle any objects could be contained in the Sequence.

    Args:
      pattern (Sequence[tuple[float,...]]): The sequence to match.

    Returns:
      Sequence[int]: the index positions at which matches are found.

    """
    j, k, = 0, 0
    finds = []
    table = self._get_table(pattern)
    while j < len(self.sequence):
      if self._equal_tuples(pattern[k], self.sequence[j]):
        j = j + 1
        k = k + 1
        if k == len(pattern):
          finds.append(j - k)
          k = table[k]
      else:
        k = table[k]
        if k < 0:
          j = j + 1
          k = k + 1
    return finds

  def _get_table(
      self,
      pattern:Sequence) -> Sequence[int]:
    """Return 'offsets' table as required by self.find_matches.

    Note that the terse variable names are those used in the KMP 1977 paper.

    Args:
      pattern (Iterable): The pattern to set up the table for.
      equals (function, optional): A function to use to test for equality of
        elements in patterns. Defaults to _equal_tuples().

    Returns:
      Iterable[int]: _description_

    """
    pos = 1
    cnd = 0
    T = {0: -1}
    while pos < len(pattern):
      if np.allclose(pattern[pos], pattern[cnd]):
        T[pos] = T[cnd]
      else:
        T[pos] = cnd
        while cnd >= 0 and not self._equal_tuples(pattern[pos], pattern[cnd]):
          cnd = T[cnd]
      pos = pos + 1
      cnd = cnd + 1
    T[pos] = cnd
    return tuple(T.values())

  def _equal_tuples(
      self,
      t1:Sequence[float],
      t2:Sequence[float]) -> bool:
    """Test for near equality of two iterables of floats.

    We use numpy.allclose to avoid issues with floating point exact equality
    tests. Wrapped to allow find_matches function to take alternative equality
    tests as input.

    Args:
      t1 (Sequence[float]): First set of floats.
      t2 (Sequence[float]): Second set of floats.

    Returns:
      bool: True if t1 == t2 for all corresponding items in each sequence.

    """
    return np.allclose(t1, t2, atol = tiling_utils.RESOLUTION * 10)

  def _round_tuple(
      self,
      t:Sequence[float],
      digits:int = 3) -> Sequence[float]:
    """Round all floats in a sequence.

    Args:
      t (Sequence[float]): the floats to round.
      digits (int): number of digits to round to.

    Returns:
      Sequence[float]: the rounded Sequence.

    """
    return tuple([np.round(x, digits) for x in t])


@dataclass(slots=True, repr=False)
class Transform:
  """Class to store details of a transform and draw it."""

  transform_type:str
  """Type of transform 'rotation', 'reflection', 'translation' or 'identity'."""
  angle:float
  """Angle of rotation (degrees)."""
  centre:geom.Point
  """Centre of the transformation."""
  translation:tuple[float,...]
  """X and Y coordinates shifts of a translation transform. A glide reflection
  may also include this."""
  transform:tuple[float,...]
  """Six element tuple for the transform in shapely.transform format. See
  https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations
  and methods in `weavingspace.tiling_utils`."""

  def __str__(self) -> str:
    if self.transform_type == "rotation":
      return (
        f"{self.transform_type:<11}{self.angle:>6.1f}° "
        f"POINT ({self.centre.x:6.1f} {self.centre.y:6.1f}) "
        f"{tuple([float(np.round(x, 3)) for x in self.transform])}"
      )
    if self.transform_type == "translation":
      return (
        f"{self.transform_type:<11}{'':>14}"
        f"({self.translation[0]:>6.1f},{self.translation[0]:>6.1f}) "
        f"{tuple([float(np.round(x, 3)) for x in self.transform])}"
      )
    if self.transform_type == "reflection":
      return (
        f"{self.transform_type:<11}{self.angle:>6.1f}° "
        f"POINT ({self.centre.x:6.1f} {self.centre.y:6.1f}) "
        f"{tuple([float(np.round(x, 3)) for x in self.transform])}"
      )
    return (
      f"{self.transform_type:<11}{'':>30}"
      f"{tuple([float(np.round(x, 3)) for x in self.transform])}"
    )


  def __repr__(self) -> str:
    return str(self)


  def apply(
      self,
      geometry:shapely.Geometry) -> shapely.Geometry:
    """Apply this transform to supplied shapely geometry and return result.

    Args:
      geometry (shapely.Geometry): a shapely geometry to transform.

    Returns:
      shapely.Geometry: the resulting transformed shapely geometry.

    """
    return affine.affine_transform(geometry, self.transform)


  def draw(
      self,
      ax:plt.Axes,
      **kwargs:str,
    ) -> plt.Axes:
    """Draw this transform on the supplied axes.

    Arguments specific to each transform type are supplied as **kwargs and are
    documented in `draw_rotation`, `draw_reflection`, and `draw_translation`.

    Args:
      ax (pyplot.axes): the axes on which to draw the transform.
      kwargs: passed on to specific draw method.

    Returns:
      pyplot.Axes: the axes with the rendering of this transform added.

    """
    match self.transform_type:
      case "rotation":
        args = inspect.signature(self.draw_rotation).parameters
        rotn_args = {k: kwargs.pop(k) for k in dict(kwargs) if k in args}
        return self.draw_rotation(ax = ax, **rotn_args)
      case "reflection":
        args = inspect.signature(self.draw_reflection).parameters
        refn_args = {k: kwargs.pop(k) for k in dict(kwargs) if k in args}
        return self.draw_reflection(ax = ax, **refn_args)
      case "translation":
        args = inspect.signature(self.draw_translation).parameters
        trans_args = {k: kwargs.pop(k) for k in dict(kwargs) if k in args}
        return self.draw_translation(ax = ax, **trans_args)
      case "identity":
        w = ax.get_window_extent().width
        return ax.set_title(f"{self.transform_type}", fontsize = w / 20)


  def draw_rotation(
      self,
      ax:plt.Axes,
      radius:float = 200,
      add_title:bool = True,
    ) -> plt.Axes:
    """Draw a rotation transform on the supplied Axes.

    Args:
      ax (pyplot.Axes): axes on which to draw the representation.
      radius (float): radius in object units to draw arcs representing the
        angle. Defaults to 200.
      add_title (bool): whether or not to add a title to the drawing.
        Defaults to True.

    Returns:
      the Axes with the rendering of this transform added.

    """
    x, y = self.centre.x, self.centre.y
    axis = geom.LineString([(x, y), (x + radius * 1.25, y)])
    arc = geom.LineString([
      geom.Point(x + radius * np.cos(np.radians(a)),
                 y + radius * np.sin(np.radians(a)))
      for a in np.linspace(0, self.angle, 50)])
    gpd.GeoSeries(self.centre).plot(
      ax = ax, color = "k", markersize = 4, zorder = 5)
    gpd.GeoSeries([axis, arc]).plot(ax = ax, color = "k", lw = .5)
    if add_title:
      w = ax.get_window_extent().width
      ax.set_title(
        f"{self.transform_type} {round(self.angle, 2)}", fontsize = w / 20)
    return ax


  def draw_reflection(
      self,
      ax:plt.Axes,
      w:float = 5,
      mirror_length:float = 100,
      add_title:bool = True,
    ) -> plt.Axes:
    """Draw a reflection transform on the supplied Axes.

    Args:
      ax (pyplot.Axes): axes on which to draw the representation.
      w (float): linewidth for the mirror line. Defaults to 5.
      mirror_length (float): length of the mirror line in units of the
        objects being reflected. Defaults to 100.
      add_title (bool): whether or not to add a title to the drawing.
        Defaults to True.

    Returns:
      the Axes with the rendering of this transform added.

    """
    x, y = self.centre.x, self.centre.y
    dx, dy = self.translation
    r = np.sqrt(self.translation[0] ** 2 + self.translation[1] ** 2)
    mirror = geom.LineString([
      (x - mirror_length / 2 * np.cos(np.radians(self.angle)),
       y - mirror_length / 2 * np.sin(np.radians(self.angle))),
      (x + mirror_length / 2 * np.cos(np.radians(self.angle)),
       y + mirror_length / 2 * np.sin(np.radians(self.angle)))])
    gpd.GeoSeries([mirror]).plot(
      ax = ax, color = "k", lw = 1, ls = "dashdot", zorder = 5)
    no_slide = np.isclose(r, 0, rtol = 1e-6, atol = 1e-6)
    if not no_slide:
      ax.arrow(
        x - dx / 2, y - dy / 2, dx, dy, length_includes_head = True,
        width = w, fc = "k", ec = None, head_width = w * 6, zorder = 5)
    if add_title:
      w = ax.get_window_extent().width
      ax.set_title(
        f"{self.transform_type} {round(self.angle, 2)}", fontsize = w / 20)
    return ax


  def draw_translation(
      self,
      ax:plt.Axes,
      c:geom.Point,
      w:float = 5,
      add_title:bool = True,
    ) -> plt.Axes:
    """Draw a translation transform on the supplied Axes.

    Args:
      ax (pyplot.Axes): axes on which to draw the representation.
      c (geom.Point): an origin point in the coordinate space of the objects
        from which to start the arrow representing the translation.
      w (float): linewidth for the arrow. Defaults to 5.
      add_title (bool): whether or not to add a title to the drawing.
        Defaults to True.

    Returns:
      the Axes with the rendering of this transform added.

    """
    gpd.GeoSeries([c]).plot(ax = ax, color = "k")
    ax.arrow(c.x, c.y, self.translation[0], self.translation[1],
             length_includes_head = True, lw = 0.5, width = w,
             fc = "k", ec = None, head_width = w * 6, zorder = 5)
    if add_title:
      w = ax.get_window_extent().width
      ax.set_title(
        (f"{self.transform_type} "
         f"({self.translation[0]:.1f}, {self.translation[1]:.1f})"),
        fontsize = w / 20)
    return ax


@dataclass(slots=True, init=False)
class Symmetries:
  """Class to identify and store symmetries of a shapely.Polygon."""

  polygon:geom.Polygon
  """Polygon from which these symmetries are derived."""
  matcher:KMP_Matcher
  """the subsequence matcher used in symmetry detection."""
  n:int
  """number of vertices of the polygon."""
  poly_code:list[tuple[float,...]]
  """the encoding of the polygon as sequence of length, angle pairs."""
  poly_code_r:list[tuple[float,...]]
  """the reversed encoding used to detect reflection symmetries."""
  rotation_shifts:list[int]
  """list of number of 2pi/n rotation symmetries."""
  reflection_shifts:list[int]
  """list of pi/n relection angle symmetries."""
  symmetries:list[Transform]
  """list of Transform objects with more complete information."""
  symmetry_group:str
  """the code denoting the symmetry group"""

  def __init__(self, polygon:geom.Polygon) -> None:
    """Construct a Symmetries instance."""
    self.polygon = polygon
    self.n = shapely.count_coordinates(self.polygon) - 1
    self.poly_code = self._get_polygon_code(self.polygon)
    self.poly_code_r = self._get_polygon_code(self.polygon, mirrored = True)
    self.matcher = KMP_Matcher(self.poly_code + self.poly_code[:-1])
    self.symmetries = self.get_symmetries()
    self.symmetry_group = self.get_symmetry_group_code()


  def _get_polygon_code(
      self,
      polygon:geom.Polygon,
      mirrored:bool = False,
    ) -> list[tuple[float,...]]:
    r"""Return list of length-angle pairs to encode a polygon shape.

    This encoding is unique up to scale. The alignment of lengths and angles is
    important and symmetry detection is sensitively dependent on it...

    The forward (i.e. unmirrored) polygon pairs are of edge lengths and the
    angle between each edge and its successor going CW around the polygon.

           A[i] ----L[i]---- A (i+1)
          /                   \
         /                     \

    i.e. edge i is between angles i and i + 1, and angle i is between edge
    i-1 and i. The mirrored encoding pairs length[i] with angle[i+1].

    The mirrored encoding pairs matched indexes of lengths and angles in the
    original polygon. This means the angle is still the one between and edge
    and its successor but proceeding CCW around the polygon.

    See in particular for clarification:

      Eades P. 1988. Symmetry Finding Algorithms. In Machine Intelligence and
      Pattern Recognition, ed. GT Toussaint, 6:41-51. Computational Morphology.
      North-Holland. doi: 10.1016/B978-0-444-70467-2.50009-6.

    Args:
      polygon (geom.Polygon): the polygon to encode
      mirrored (bool, optional): if true encoding will be in CCW order.
        Defaults to False.

    Returns:
      list[tuple[float]]: a list of side-length angle pairs.

    """
    # get edge lengths and angles
    lengths = tiling_utils.get_side_lengths(polygon)
    raw_angles = tiling_utils.get_interior_angles(polygon)
    if mirrored:
      lengths_r = lengths[::-1]
      angles_r = raw_angles[::-1]
      return list(zip(lengths_r, angles_r, strict = True))
    angles = raw_angles[1:] + raw_angles[:1]
    return list(zip(lengths, angles, strict = True))


  def get_symmetries(self) -> list[Transform]:
    """Find rotation and reflection symmetries of a polygon.

    The work is delegated to `get_rotations()` and `get_reflections()`.

    Based on

      Eades P. 1988. Symmetry Finding Algorithms. In Machine Intelligence and
      Pattern Recognition, ed. GT Toussaint, 6:41-51. Computational Morphology.
      North-Holland. doi: 10.1016/B978-0-444-70467-2.50009-6.

    and also

      Wolter JD, TC Woo, and RA Volz. 1985. Optimal algorithms for symmetry
      detection in two and three dimensions. The Visual Computer 1(1): 37-48.
      doi: 10.1007/BF01901268.

    Details in these papers are not unambiguous. This implementation was
    developed based on them, but with some trial and error to get the index
    offsets and (especially) retrieval of the reflection axes angles to work.

    Returns:
      list[Transform]: a list of Transform objects representing the polygon
        symmetries.

    """
    return self.get_rotations() + self.get_reflections()


  def get_rotations(
      self,
      offsets:None|list[int] = None,
    ) -> list[Transform]:
    """Get the rotations associated with this collection of symmetries.

    Returns:
      list[Transform]: A list of the rotation symmetries associated with this
        Symmetries instance's polygon.

    """
    if offsets is None:
      self.rotation_shifts = self.matcher.find_matches(self.poly_code)
      offsets = self.rotation_shifts
    c = self.polygon.centroid
    rot_angles = [np.round(i * 360 / self.n, 6) for i in offsets]
    rotation_transforms = [tuple(
      np.round(tiling_utils.get_rotation_transform(a, (c.x, c.y)), 6))
      for a in rot_angles]
    return [
      Transform("rotation", angle, c, (0, 0), transform) for
      transform, angle in zip(rotation_transforms, rot_angles, strict = True)]


  def get_reflections(
      self,
      offsets:None|list[int] = None,
    ) -> list[Transform]:
    """Get the reflections associated with this collection of symmetries.

    Args:
      offsets (list[int]): list of integer offsets i where i / n * 360 where n
        is the number of sides of this Symmetries instance's polygons is the
        angle of the mirror in which a reflection symmetry has been identified.

    Returns:
      list[Transform]: A list of the reflection symmetries associated with this
        Symmetries instance's polygon.

    """
    # collate all possible reflection axes - ordering of these is important
    # for correct picking - so don't mess with this code!
    bearings = tiling_utils.get_side_bearings(self.polygon)
    reflection_axes = []
    interior_angles = tiling_utils.get_interior_angles(self.polygon)
    if offsets is None:
      self.reflection_shifts = self.matcher.find_matches(self.poly_code_r)
      offsets = self.reflection_shifts
    for i in range(self.n):
      # the angle bisectors
      reflection_axes.append(bearings[i] - interior_angles[i] / 2)
      # the edge_perpendicular bisectors
      reflection_axes.append(bearings[i] - 90)  # normal to the edge
    # pick out those where matches occurred using the offsets
    ref_angles = [np.round(reflection_axes[i], 6) for i in offsets]
    c = self.polygon.centroid
    reflection_transforms = [tuple(
      np.round(tiling_utils.get_reflection_transform(angle, (c.x, c.y)), 6))
      for angle in ref_angles]
    return [
      Transform("reflection", angle, c, (0, 0), transform) for
      transform, angle in zip(reflection_transforms, ref_angles, strict = True)]

  def get_symmetry_group_code(self) -> str:
    """Return symmetry group code for this instance's polygon.

    See https://en.wikipedia.org/wiki/Symmetry_group#Two_dimensions.

    Returns:
        str: code such as C2 or D4 for cyclic or dihedral symmetry groups.

    """
    if len(self.reflection_shifts) == 0:
      # no reflection symmetries, so it is a cyclic group Cn
      return f"C{len(self.rotation_shifts)}"
    # must be the dihedral group Dn
    return f"D{len(self.rotation_shifts)}"


  def get_corner_offset(
      self,
      poly2:geom.Polygon) -> int|None:
    """Find offset between corners of this polygon and another one.

    The offset is determined under whatever transformation will match one on
    to the other. Used by the `Topology` class. The offset is expressed as the
    number of corners clockwise by which the supplied polygon is offset from
    this one.

    Args:
      poly2 (geom.Polygon): polygon for which the offset is desired.

    Returns:
      int|None: integer number of corners clockwise by which the polygon
        is offset from this one, or None if they don't match.

    """
    s2 = Symmetries(poly2)
    matches = self.matcher.find_matches(s2.poly_code)
    if len(matches) > 0:
      return matches[0]
    matches = self.matcher.find_matches(s2.poly_code_r)
    if len(matches) > 0:
      return -matches[0]
    return None


@dataclass(slots=True, init=False)
class ShapeMatcher:
  """Class to manage finding transforms that map one polygon on to another.

  This was previously handled 'internally' by an earlier version of the
  Symmetries code, but that quickly became ungainly and difficult to maintain.
  This approach makes for cleaner code, and does the additional work of finding
  e.g. centres of rotations (which may lie outside either or both polygons).
  """

  shape:geom.Polygon
  """The target polygon on to which matching transforms are required."""
  s1:Symmetries
  """The Symmetries of the target polygon."""

  def __init__(  # noqa: D107
      self,
      shape: geom.Polygon) -> None:
    self.shape = shape
    self.s1 = Symmetries(shape)


  def get_polygon_matches(
      self,
      shape2:geom.Polygon) -> list[Transform]|None:
    """Return all transforms that match shape2 onto this instance's polygon.

    Args:
      shape2 (geom.Polygon): polygon for which matching transforms are
        requested.

    Returns:
      list[Transform]|None: a list of Transforms that will match shape2
        onto this instance's shape. None is returned if no matching transforms
        are found.

    """
    s2 = Symmetries(shape2)
    if self.s1.symmetry_group != s2.symmetry_group:
      return None
    match_rot, rots = self._get_rotation_matches(s2)
    match_ref, refs = self._get_reflection_matches(s2)
    if match_rot and match_ref:
      return rots + refs
    if match_rot:
      return rots
    if match_ref:
      return refs
    return None


  def _get_rotation_matches(
      self,
      s2:Symmetries) -> tuple[bool,list[Transform]]:
    """Get rotations to match s2's polygon onto this instance's polygon.

    Args:
      s2 (Symmetries): Symmetries associated with the polygon of interest.

    Returns:
      tuple[bool,list[Transform]]: boolean is True if there are any matches
        when matching transforms will be included in the associated list. If
        there are no matches (False, []) will be returned.

    """
    # need this locally to check for equality of all members of a list
    def all_equal(lst:list[float]) -> bool:
      return all(np.isclose(lst[0], x, 1e-3, 1e-3) for x in lst)

    matches = self.s1.matcher.find_matches(s2.poly_code)
    if len(matches) == 0:
      return False, None
    transforms = []
    # get lists of polygon corners aligned correctly to measure the angle
    p1_corners = tiling_utils.get_corners(self.shape, repeat_first = False)
    p2_corners = tiling_utils.get_corners(s2.polygon, repeat_first = False)
    corner_distances = self._get_corner_distances_between_polys(
      p1_corners, p2_corners, matches)
    equalities = [all_equal(d) for d in corner_distances]
    if all(equalities):
      # the two shapes are identical and in the same spot
      # so rotations symmetries are those of the target shape
      transforms.extend(self.s1.get_rotations(matches))
      return True, transforms
    if any(equalities):
      # the one where all are equal is a translation so we'll skip it below
      # and add it here as a translation transform
      m = equalities.index(True)
      del matches[m]
      dx, dy = (p1_corners[0].x - p2_corners[m].x,
                p1_corners[0].y - p2_corners[m].y)
      transforms.append(
        Transform("translation", None, None, (dx, dy), (1, 0, 0, 1, dx, dy)))
    for m in matches:
      a, c = self._get_angle_between_polygons(p1_corners, p2_corners, m)
      if a is None:
        continue
      transforms.append(Transform("rotation", a, c, (0, 0),
        tiling_utils.get_rotation_transform(a, (c.x, c.y))))
    return True, transforms


  def _get_corner_distances_between_polys(
      self,
      corners1:list[geom.Point],
      corners2:list[geom.Point],
      offsets:list[int],
    ) -> list[list[float]]:
    return [
      [p1.distance(p2) for p1, p2 in
       zip(corners1, corners2[i:] + corners2[:i], strict = True)]
      for i in offsets]


  def _get_angle_between_polygons(
      self,
      corners1:list[geom.Point],
      corners2:list[geom.Point],
      offset:int,
    ) -> tuple[None,...]|tuple[str,tuple[float,...]]|tuple[float,geom.Point]:
    """Return angle by which lists of corners are rotationally offset.

    If lists are at the same orientation then the translation required to match
    them (if any) will be returned.

    Args:
      corners1 (list[geom.Point]): list of corners of the first polygon.
      corners2 (list[geom.Point]): list of corners of the second polygon.
      offset (int): number of clockwise steps through the corners by which
        corners2 are offset from corners1 as previously determined.

    Returns:
      tuple[float,geom.Point]|tuple[str,tuple[float]]: one of
      ('identity' (0, 0)), ('translation', (x, y)), or (angle, geom.Point),
        where the last of these is an angle and a centre of rotation.

    """
    corners2 = corners2[offset:] + corners2[:offset]
    dists = [
      p1.distance(p2) for p1, p2 in zip(corners1, corners2, strict = True)]
    if all(np.isclose(dists[0], d, atol = 1e-3, rtol = 1e-3) for d in dists):
      # polygons already aligned similarly so we just need to find translation
      if np.isclose(dists[0], 0, atol = 1e-3, rtol  = 1e-3):
        return "identity", (0, 0)
      if offset == 0:
        return "translation", (corners1[0].x - corners2[0].x,
                               corners1[0].y - corners2[0].y)
    # Polygons not lined up. Order them by distances between them
    ordered_dists = sorted([(i, d) for i, d in enumerate(dists)],
                           key = lambda x: x[1], reverse = True)
    # the two furthest will give the most reliable numbers
    a, b = [i for i, d in ordered_dists[:2]]
    p_1a, p_1b = corners1[a], corners1[b]
    p_2a, p_2b = corners2[a], corners2[b]
    # get the perpendiculars of the lines joining A and B in each polygon
    perp_1a_2a = tiling_utils.get_straight_line(
      p_1a, p_2a, perpendicular = True)
    perp_1b_2b = tiling_utils.get_straight_line(
      p_1b, p_2b, perpendicular = True)
    # and intersect them
    centre = tiling_utils.get_intersection(perp_1a_2a, perp_1b_2b)
    if centre is None:
      return None, None
    angle = -tiling_utils.get_inner_angle(p_1a, centre, p_2a)
    if angle < -179:
      angle = angle + 360
    elif angle > 180:
      angle = angle - 360
    return angle, centre


  def _get_reflection_matches(
      self,
      s2:Symmetries,
    ) -> tuple[bool,list[Transform]|None]:
    """Get reflections that match s2's polygon to this instance's polygon.

    Args:
      s2 (Symmetries): Symmetries associated with the polygon of interest.

    Returns:
      tuple[bool,list[Transform]]: boolean is True if there are any matches
        when matching transforms will be included in the associated list. If
        there are no matches (False, []) will be returned.

    """
    matches = self.s1.matcher.find_matches(s2.poly_code_r)
    if len(matches) == 0:
      return False, None
    # get mean centroid - all reflection matches will pass through this
    c1 = self.s1.polygon.centroid
    c2 = s2.polygon.centroid
    c = ((c1.x + c2.x) / 2, (c1.y + c2.y) / 2)
    # get each polygon's reflection symmetries
    reflections1 = self.s1.get_reflections(matches)
    reflections2 = s2.get_reflections(matches)
    # their mean angle will be a line of reflection through c
    angles = [(ref1.angle + ref2.angle) / 2
              for ref1, ref2 in zip(reflections1, reflections2, strict = True)]
    transforms = [tiling_utils.get_reflection_transform(angle, c)
                  for angle in angles]
    # there may also be a 'glide' component additional to the reflection
    # to find it reflect the centre of polygon 2 by each transform and
    # then find translation necessary to map it onto the centre of polygon 1
    ctrs2r = [affine.affine_transform(c2, transform)
              for transform in transforms]
    dxdys = [(c1.x - ctr2r.x, c1.y - ctr2r.y) for ctr2r in ctrs2r]
    return (True,
      [Transform(
        "reflection", angle, geom.Point(c), dxdy,
        tiling_utils.combine_transforms([
          tiling_utils.get_reflection_transform(angle, c),
          tiling_utils.get_translation_transform(dxdy[0], dxdy[1])]))
          for angle, dxdy in zip(angles, dxdys, strict = True)])



"""Handles combining the ordering of which strands are on top in weaves."""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np

"""Initially in a biaxial weave intersection sites are encoded numerically:

  1 = warp is absent
  2 = weft is absent
  3 = both threads are absent
  4 = weft is on top
  5 = warp is on top

this tuple of dictionaries translates these to tuples listing the 'layer' order
depending on the input axis, where item 0 is the biaxial case with layers 0 and
1 only, item 1 is layers 0 and 1, item 2 is layers 1 and 2, and item 3 is
layers 2 and 0.
"""
_decode_orders:tuple[dict,dict,dict,dict] = (
  {1: (0,  ), 2: (1,  ), 3: (   ), 4: (0, 1), 5: (1, 0)}, # biaxial case
  {1: (1,  ), 2: (0,  ), 3: (   ), 4: (1, 0), 5: (0, 1)}, # triaxial 0|1
  {1: (2,  ), 2: (1,  ), 3: (   ), 4: (2, 1), 5: (1, 2)}, # triaxial 1|2
  {1: (0,  ), 2: (2,  ), 3: (   ), 4: (0, 2), 5: (2, 0)}, # triaxial 2|0
)

"""Nested dictionary to combine layer orders from 3 biaxial weaves
into a single consistent order of layers. Inconsistent combinations
will not return a value. Layers 0 and 1 are present in the first 'tier',
layers 1 and 2 in the second and 2 and 0 in the third.
"""
_combined_orderings:dict[tuple,dict] = {
  (    ): {(    ): {(    ): None,         # -- -- --  ->  All absent
                    (2,  ): (2,  )},      # -- -- 2-  ->  Only 2 present
           (1,  ): {(    ): (1,  )},      # -- 1- --  ->  Only 1 present
           (2,  ): {(2,  ): (2,  )}},     # -- 2- 2-  ->  Only 2 present
  (0,  ): {(    ): {(0,  ): (0,  ),       # 0- -- 0-  ->  Only 0 present
                    (    ): (0,  )},      # 0- -- --  ->  Only 0 present
           (2,  ): {(2, 0): (2, 0),       # 0- 2- 20  ->  2 > 0
                    (0, 2): (0, 2)}},     # 0- 2- 02  ->  0 > 2
  (1,  ): {(1,  ): {(    ): (1,  )},      # 1- 1- --  ->  Only 1 present
           (1, 2): {(2,  ): (1, 2)},      # 1- 12 2-  ->  1 > 2
           (2, 1): {(2,  ): (2, 1)}},     # 1- 21 2-  ->  2 > 1
  (0, 1): {(1,  ): {(0,  ): (0, 1)},      # 01 1- 0-  ->  0 > 1
           (1, 2): {(0, 2): (0, 1, 2)},   # 01 12 02  ->  0 > 1 > 2
           (2, 1): {(2, 0): (2, 0, 1),    # 01 21 20  ->  2 > 0 > 1
                    (0, 2): (0, 2, 1)}},  # 01 21 02  ->  0 > 2 > 1
  (1, 0): {(1,  ): {(0,  ): (1, 0)},      # 10 1- 0-  ->  1 > 0
           (1, 2): {(2, 0): (1, 2, 0),    # 10 12 20  ->  1 > 2 > 0
                    (0, 2): (1, 0, 2)},   # 10 12 02  ->  1 > 0 > 2
           (2, 1): {(2, 0): (2, 1, 0)}},  # 10 21 20  ->  2 > 1 > 0
}


@dataclass(slots=True)
class Loom:
  """A collection of weave matrices and associated attributes.

  Central to the loom object are the indices and orderings lists. These
  list the grid coordinate pairs and corresponding layer orderings at the
  site indexed by the coordinates. E.g., for a simple plain weave, we have:

    indices:   [(0, 0), (0, 1), (1, 0), (1, 1)]
    orderings: [(0, 1), (1, 0), (1, 0), (0, 1)]
  """

  indices:list[tuple]
  """grid coordinate pairs."""
  orderings:list
  """list of layer orders at site locations index by corresponding times in
  `indices`."""
  dimensions:tuple
  """maximum coordinate values in each direction."""
  orientations:tuple
  """angles (in degrees) which strands on each axis make with the x-axis.
  (0, -90) for 2 axes, (0, 120, 240) for 3."""
  n_axes:int
  """number of axes, 2 or 3."""

  def __init__(
      self,
      *matrices:np.ndarray,
    ) -> None:
    """Return a Loom instance.

    Takes either one or three weave matrices as input and initialises the loom
    based on these.
    """
    if len(matrices) == 1:
      m = matrices[0]
      self.dimensions = m.shape
      self.n_axes = len(self.dimensions)
      self.orientations = (0, -90)
      self.indices = [(i, j) for i in range(m.shape[0])
                             for j in range(m.shape[1])]
      self.orderings = [_decode_orders[0][m[ij]] for ij in self.indices]
    else:
      n_01, n_12, n_20 = [max(m.shape) for m in matrices]
      self.dimensions = (n_01, n_12, n_20)
      self.n_axes = len(self.dimensions)
      self.orientations = (0, 120, 240)
      all_indices = [(i, j, k) for i in range(n_01)
                               for j in range(n_12)
                               for k in range(n_20)]
      # parity is used to select coordinate triples in the grid see:
      # Nagy BN 2003. Shortest Paths in Triangular Grids with
      # Neighbourhood Sequences. Journal of Computing and Information
      # Technology 11 (2):111 https://doi.org/10.2498/cit.2003.02.04
      parity = (n_01 + n_12 + n_20 - 3) // 2
      self.indices = [x for x in all_indices if sum(x) in (parity, parity + 1)]
      m1, m2, m3 = matrices
      # extend the input 2D grids if needed
      m_01 = np.tile(m1, (np.lcm(n_01, m1.shape[0]) // m1.shape[0],
                          np.lcm(n_01, m1.shape[1]) // m1.shape[1]))
      m_12 = np.tile(m2, (np.lcm(n_12, m2.shape[0]) // m2.shape[0],
                          np.lcm(n_12, m2.shape[1]) // m2.shape[1]))
      m_20 = np.tile(m3, (np.lcm(n_20, m3.shape[0]) // m3.shape[0],
                          np.lcm(n_20, m3.shape[1]) // m3.shape[1]))
      # get the layer orders in each using 2 of the 3 coordinates
      order_01 = [_decode_orders[1][m_01[ij]]
                  for ij in [(x[1], x[0]) for x in self.indices]]
      order_12 = [_decode_orders[2][m_12[ij]]
                  for ij in [(x[2], x[1]) for x in self.indices]]
      order_20 = [_decode_orders[3][m_20[ij]]
                  for ij in [(x[0], x[2]) for x in self.indices]]
      # # combine orders from the three matrices stacked
      self.orderings = [
        self._combine_orders(abc)
        for abc in zip(order_01, order_12, order_20, strict = True)]


  # convenience wrapper for the combined_orderings dictionary
  # missing values return "NA"
  def _combine_orders(
      self,
      orders:tuple[tuple[int],...],
    ) -> tuple[int]|str:
    try:
      result = _combined_orderings[orders[0]][orders[1]][orders[2]]
    except KeyError as e:
      e.add_note(f"Unable to determine unique ordering on {orders}")
      return "NA"
    else:
      return result



"""Utility functions for the `weavingspace` module.

Most are geometric convenience functions for commonly applied
operations.
"""

from __future__ import annotations

import re
import string
from typing import TYPE_CHECKING, NamedTuple

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.affinity as affine
import shapely.geometry as geom
import shapely.ops
from shapely.algorithms import polylabel

if TYPE_CHECKING:
  from collections.abc import Iterable

PRECISION = 6
RESOLUTION = 1e-6

def _parse_strand_label(s:str) -> list[str]:
  """Break a strand label specification in to a list of labels.

  Args:
    s (str): see get_strand_ids() for details

  Returns:
    list[str]: list of strand labels.

  """
  # replace multiple open and close parentheses with single parentheses
  clean_s = re.sub("[(]+", "(", re.sub("[)]+", ")", s))
  is_combo = False
  output = []
  current = ""
  for c in clean_s:
    if is_combo:
      if c == ")":
        # end of a combination
        output.append(current)
        current = ""
        is_combo = False
      else:
        current = current + c
    elif c == "(":
      # start of a combination
      is_combo = True
    else:
      output.append(c)
  return output


def get_strand_ids(strands_spec: str) -> tuple[list[str]]:
  """Convert strands specification string to list of lists of strand labels.

  String format is "a|bc|(de)f" where | separates strands in each
  direction and () designates combining labels into a single strand that
  will be sliced lengthwise. Example output:

        "a|bc|(de)f" -> (["a"], ["b", "c"], ["de", "f"])

  Superflous parentheses are removed, but no other error-checks are
  applied.

  Args:
    strands_spec (str): the strands specification to be parsed.

  Returns:
    tuple[str]: tuple of lists of labels for each set of strands.

  """
  strand_ids = [_parse_strand_label(s) for s in strands_spec.split("|")]
  strand_ids = strand_ids if len(strand_ids) == 3 else [*strand_ids, [""]]
  return tuple(strand_ids)


def get_regular_polygon(spacing:float, n:int) -> geom.Polygon:
  r"""Return regular n-gon centered on (0, 0) with height specified.

  n-gon has horizontal base and height given by spacing. Generation is based on
  equally spaced points around a circle with a radius calculated slightly
  differently, depending on whether the number of sides is even or odd.

            Even         Odd
        \  /    \  /      |
         \/   ___\/___    |
         /\      /\      / \
        /  \    /  \    /   \

  The returned polygon is gridified. Note also that shapely silently reorders
  the coordinates clockwise from the lower left corner, regardless of the CCW
  order in which we supply them.

  Args:
    spacing (_type_): required height.
    n (int): number of sides.

  Returns:
    geom.Polygon: required geom.Polygon.

  """
  if n % 2 == 0:
    R = spacing / np.cos(np.pi / n) / 2
  else:
    R = spacing / (1 + np.cos(np.pi / n))
  angle_0 = -np.pi / 2 + np.pi / n
  a_diff = 2 * np.pi / n
  angles = [angle_0 + a_diff * i for i in range(n)]
  corners = [(R * np.cos(angle), R * np.sin(angle)) for angle in angles]
  return geom.Polygon(corners)


def get_corners(shape:geom.Polygon,
                repeat_first:bool = True) -> list[geom.Point]:
  """Return list of geom.Points on boundary of a polygon.

  The first point is optionally repeated. No simplification is carried out
  (e.g. if a line segment has a 'corner' along its length, it is NOT removed;
  see get_clean_polygon for that). Points have precision set to the
  packagedefault tiling_utils.RESOLUTION.

  Args:
    shape (geom.Polygon): polygon whose corners are required.
    repeat_first (bool, optional): if True the first corner is repeated in the
      returned list, if False it is omitted. Defaults to True.

  Returns:
    list[geom.Point]: list of geom.Point vertices of the polygon.

  """
  corners = [geom.Point(pt) for pt in shape.exterior.coords]
  return corners if repeat_first else corners[:-1]


def get_sides(shape:geom.Polygon) -> list[geom.LineString]:
  """Return polygon sides as list of geom.LineStrings.

  Resolution is set to the package default tiling_utils.RESOLUTION. No
  simplification for successive colinear sides is carried out.

  Args:
    shape (geom.Polygon): polygon whose edges are required.

  Returns:
    list[geom.LineString]: list of geom.LineString sides of the polygon.

  """
  corners = get_corners(shape)
  return [geom.LineString([p1, p2])
          for p1, p2 in zip(corners[:-1], corners[1:], strict = True)]


def get_side_lengths(shape:geom.Polygon) -> list[float]:
  """Return list of lengths of polygon sides.

  No simplification for corners along sides is carried out.

  Args:
    shape (geom.Polygon): polygon whose edge lengths are required.

  Returns:
    list[float]: list of side lengths.

  """
  corners = get_corners(shape)
  return [
    p1.distance(p2) for p1, p2 in zip(corners[:-1], corners[1:], strict = True)]


def get_side_bearings(shape:geom.Polygon) -> tuple[float,...]:
  """Return list of angles between sides of polygon and the positive x-axis.

  Angle is in degrees and calculated when proceeding from the first point in
  each side to its end point. This should usually be CW around the polygon.

  Args:
    shape (geom.Polygon): polygon whose side bearings are required.

  Returns:
    tuple[float,...]: tuple of bearings of each edge.

  """
  return tuple([np.degrees(np.arctan2(
    e.coords[1][1] - e.coords[0][1], e.coords[1][0] - e.coords[0][0]))
    for e in get_sides(shape)])


def get_interior_angles(shape:geom.Polygon) -> list[float]:
  """Return angles (in degrees) between successive edges of a polygon.

  No polygon simplification is carried out so some angles may be 180 (i.e. a
  'corner' along a side, such that successive sides are colinear).

  Args:
    shape (geom.Polygon): polygon whose angles are required.

  Returns:
    list[float]: list of angles.

  """
  corners = get_corners(shape, repeat_first = False)
  wrap_corners = corners[-1:] + corners + corners[:1]
  triples = zip(wrap_corners[:-2],
                wrap_corners[1:-1],
                wrap_corners[2:], strict = True)
  return [get_inner_angle(p1, p2, p3) for p1, p2, p3 in triples]


def get_clean_polygon(
      shape:geom.MultiPolygon|geom.Polygon|gpd.GeoSeries,
    ) ->  geom.MultiPolygon|geom.Polygon|gpd.GeoSeries:
  """Return polygon with successive co-linear or very close corners removed.

  Particularly useful for tidying polygons weave tilings that have been
  assembled from multiple 'cells' in the weave grid.

  Args:
    shape (geom.MultiPolygon|geom.Polygon): polygon to clean.

  Returns:
    geom.Polygon: cleaned polygon.

  """
  if isinstance(shape, geom.MultiPolygon):
    return geom.MultiPolygon([get_clean_polygon(p) for p in shape.geoms])
  if isinstance(shape, geom.GeometryCollection):
    elements = [p for p in shape.geoms if isinstance(p, geom.Polygon)]
    if len(elements) > 1:
      return get_clean_polygon(geom.MultiPolygon(elements))
    return get_clean_polygon(elements[0])
  corners = get_corners(shape, repeat_first = False)
  # first remove any 'near neighbour' corners
  distances = get_side_lengths(shape)
  to_remove = [np.isclose(d, 0, rtol = RESOLUTION, atol = 10 * RESOLUTION)
                for d in distances]
  corners = [c for c, r in zip(corners, to_remove, strict = True) if not r]
  # next remove any that are colinear
  p = geom.Polygon(corners)
  corners = get_corners(p, repeat_first = False)
  angles = get_interior_angles(p)
  to_remove = [np.isclose(a,   0, rtol = RESOLUTION, atol = RESOLUTION) or
               np.isclose(a, 180, rtol = RESOLUTION, atol = RESOLUTION)
               for a in angles]
  corners = [c for c, r in zip(corners, to_remove, strict = True) if not r]
  return gridify(geom.Polygon(corners))


def rotate_preserving_order(
    polygon:geom.Polygon,
    angle:float,
    centre:geom.Point = None,
  ) -> geom.Polygon:
  """Return supplied polygon rotated with order of corners preserved.

  Order of polygon corners is not guaranteed to be preserved by shapely.
  affinity.rotate).

  Args:
    polygon (geom.Polygon): polygon to rotate.
    angle (float): desired angle of rotation (in degrees).
    centre (geom.Point): the rotation centre (passed on to shapely.affinity.
      rotate).

  Returns:
    geom.Polygon: rotated polygon.

  """
  ctr = "center" if centre is None else centre
  corners = get_corners(polygon, repeat_first = False)
  return geom.Polygon([affine.rotate(c, angle, ctr) for c in corners])


def ensure_cw(shape:geom.Polygon) -> geom.Polygon:
  """Return the polygon with its outer boundary vertices in clockwise order.

  It is important to note that shapely.set_precision() imposes clockwise order
  on polygons, and since it is used widely throughout theses modules, it makes
  sense to impose this order.

  Args:
    shape (geom.Polygon): the polygon.

  Returns:
    geom.Polygon: the polygon in clockwise order.

  """
  if geom.LinearRing(shape.exterior.coords).is_ccw:
    return shape.reverse()
  return shape


def get_inner_angle(p1:geom.Point, p2:geom.Point, p3:geom.Point) -> float:
  r"""Return angle between line p1-p2 and p2-p3, i.e., the angle A below.

            p2
           / \
          / A \
        p1     p3

  Args:
    p1 (geom.Point): first point.
    p2 (geom.Point): second 'corner' point.
    p3 (geom.Point): third point.

  Returns:
    float: angle in degrees.

  """
  return np.degrees(np.arctan2(p3.y - p2.y, p3.x - p2.x) -
                    np.arctan2(p1.y - p2.y, p1.x - p2.x)) % 360


def is_regular_polygon(shape:geom.Polygon) -> bool:
  """Test if supplied polygon is regular (i.e. equal sides and angles).

  Args:
    shape (geom.Polygon): polygon to test.

  Returns:
    bool: True if polygon is regular, False if not.

  """
  side_lengths = get_side_lengths(shape)
  angles = get_interior_angles(shape)
  return all(np.isclose(side_lengths, side_lengths[0])) \
     and all(np.isclose(angles, angles[0]))


def is_tangential(shape:geom.Polygon) -> bool:
  """Determine if supplied polygon is tangential.

  A tangential polygon can have a circle inscribed tangential to all its sides.

  NOTE: not currently used but retained in case incentre is restored to use
  'home made' code rather than polylabel.polylabel.

  Note that this will fail for polygons with successive colinear sides, meaning
  that polygons should be fully simplified...

  Args:
    shape (geom.Polygon): polygon to test.

  Returns:
    bool: True if polygon is tangential, False if not.

  """
  if is_regular_polygon(shape):
    return True
  side_lengths = get_side_lengths(shape)
  n = len(side_lengths)
  if n % 2 == 1:
    if n == 3: #triangles are easy!
      return True
    # odd number of sides there is a nice solvable system  of equations see
    # https://math.stackexchange.com/questions/4065370/tangential-polygons-conditions-on-edge-lengths
    #
    # TODO: this is not reliable for reasons I don't fully understand. There is
    # a corresponding crude catch exception in incentre... but this needs a
    # more complete investigation
    mat = np.identity(n) + np.roll(np.identity(n), 1, axis = 1)
    fractions = np.linalg.inv(mat) @ side_lengths / side_lengths
    if not(np.isclose(np.mean(fractions),
                      0.5, rtol= RESOLUTION, atol = RESOLUTION)):
      return False
    ones = [np.isclose(f, 1, rtol = RESOLUTION, atol = RESOLUTION)
            for f in fractions]
    zeros = [np.isclose(f, 0, rtol = RESOLUTION, atol = RESOLUTION)
             for f in fractions]
    negatives = [f < 0 for f in fractions]
    return not (any(ones) or any(zeros) or any(negatives))
  if n == 4:
    # for quadrilaterals odd and even side lengths are equal
    return np.isclose(sum(side_lengths[0::2]),
                      sum(side_lengths[1::2]), rtol = RESOLUTION)
  # other even numbers of sides... hard to avoid brute force
  bisectors = [get_angle_bisector(shape, i) for i in range(n)]
  intersects = [b1.intersection(b2) for b1, b2 in
                zip(bisectors, bisectors[1:] + bisectors[:1], strict = False)
                if b1.intersects(b2)]
  distances = [i1.distance(i2) for i1, i2 in
               zip(intersects, intersects[1:] + intersects[:1],
                   strict = False)]
  return all([d <= 2 * RESOLUTION for d in distances])  # noqa: C419


def get_incentre(shape:geom.Polygon) -> geom.Point:
  """Get a polygon incentre.

  See https://en.wikipedia.org/wiki/Incenter.

  This method relies on the polygon being tangential, i.e. there is an
  inscribed circle to which all sides of the polygon are tangent. It will work
  on all the polygons encountered in the Laves tilings, but is not guaranteed
  to work on all polygons.

  Given that the polygon is tangential, the radius of the inscribed circle is
  the [apothem of the polygon](https://en.wikipedia.org/wiki/Apothem) given by
  2 x Area / Perimeter. We apply a parallel offset of this size to two sides of
  the polygon and find their intersection to determine the centre of the circle
  givng the incentre of the polygon.

  Args:
    shape (geom.Polygon): the polygon.

  Returns:
    geom.Point: the incentre of the polygon.

  """
  if is_regular_polygon(shape):
    return shape.centroid
  return polylabel.polylabel(shape)


def get_incircle(shape:geom.Polygon) -> geom.Polygon:
  """Return incircle of supplied polygon.

  Args:
    shape(geom.Polygon): polygon for which incircle is required.

  Returns:
    geom.Polygon: a polygon representation of the incircle

  """
  c = get_incentre(shape)
  r = min([c.distance(e) for e in get_sides(shape)])
  return c.buffer(r)


def get_apothem(shape:geom.Polygon) -> float:
  """Return length of the apothem of supplied polygon.

  Args:
    shape(geom.Polygon): polygon for which apothem length is required.

  Returns:
    float: an estimate of the apothem length.

  """
  if is_regular_polygon(shape):
    return 2 * shape.area / geom.LineString(shape.exterior.coords).length
  c = polylabel.polylabel(shape, 1)
  return min([c.distance(e) for e in get_sides(shape)])


def get_angle_bisector(shape:geom.Polygon, v = 0) -> geom.LineString:
  """Return angle bisector of specified corner of supplied polygon.

  Args:
    shape (geom.Polygon): the polygon
    v (int, optional): index of the corner whose bisector is required.
    Defaults to 0.

  Returns:
    geom.LineString: line which bisects the specified corner.

  """
  pts = get_corners(shape, repeat_first = False)
  n = len(pts)
  p0 = pts[v % n]
  p1 = pts[(v + 1) % n]
  p2 = pts[(v - 1) % n]
  d01 = p0.distance(p1)
  d02 = p0.distance(p2)
  if d01 < d02:
    p2 = geom.LineString([p0, p2]).interpolate(d01)
  else:
    p1 = geom.LineString([p0, p1]).interpolate(d02)
  m = geom.Point((p1.x + p2.x) / 2, (p1.y + p2.y) / 2)
  # we have to scale this to much larger in case of an angle near 180
  return affine.scale(geom.LineString([p0, m]), 10, 10, origin = p0)


def _get_interior_vertices(tiles:gpd.GeoDataFrame) -> gpd.GeoSeries:
  """Return points not on outer boundary of supplied set of tiles.

  Args:
    tiles (gpd.GeoDataFrame): a set of polygons.

  Returns:
    gpd.GeoSeries: interior vertices of the set of polygons.

  """
  tiles:gpd.GeoSeries = gridify(tiles.geometry)
  uu:geom.Polygon = safe_union(tiles, as_polygon = True)
  interior_pts = set()
  for tile in tiles:
    for corner in tile.exterior.coords:
      if uu.contains(geom.Point(corner).buffer(1e-3, cap_style = "square")):
        interior_pts.add(corner)
  return gridify(gpd.GeoSeries([geom.Point(p) for p in interior_pts]))


def gridify(
    gs:gpd.GeoSeries | gpd.GeoDataFrame | shapely.Geometry,
  ) -> gpd.GeoSeries | gpd.GeoDataFrame | shapely.Geometry:
  """Return supplied GeoSeries rounded to tiling_utils.RESOLUTION.

  Args:
    gs (gpd.GeoSeries): geometries to gridify.
    precision (int, optional): digits of precision. Defaults to 6.

  Returns:
    gpd.GeoSeries: the rounded geometries.

  """
  if isinstance(gs, shapely.Geometry) :
    return shapely.set_precision(gs, grid_size = RESOLUTION)
  if isinstance(gs, gpd.GeoSeries):
    return gpd.GeoSeries(
      [shapely.set_precision(g, grid_size = RESOLUTION) for g in list(gs)],
      crs = gs.crs)
  if isinstance(gs, gpd.GeoDataFrame):
    gs.geometry = gpd.GeoSeries(
      [shapely.set_precision(g, grid_size = RESOLUTION)
       for g in list(gs.geometry)], crs = gs.crs)
    return gs
  return None


def get_dual_tile_unit(unit:TileUnit) -> gpd.GeoDataFrame:
  """Convert supplied TileUnit to candidate GeoDataFrame of its dual TileUnit.

  NOTE: this is complicated and not remotely guaranteed to work!
  a particular issue is that where to place the vertices of the faces
  of the dual with respect to the tiles in the original is ill-defined.
  This is because the dual process is topologically not metrically defined,
  so that exact vertex locations are ambiguous. Tiling duality is defined in
  Section 4.2 of Grunbaum B, Shephard G C, 1987 _Tilings and Patterns_ (W. H.
  Freeman and Company, New York)

  NOTE: In general, this method will work only if all supplied tiles are
  regular polygons. A known exception is if the only non-regular polygons are
  triangles.

  NOTE: 'clean' polygons are required. If supplied polygons have messy
  vertices with multiple points where there is only one proper point, bad
  things are likely to happen! Consider using `clean_polygon()` on the
  tile geometries.

  Because of the above limitations, we only return a GeoDataFrame
  for inspection. However some `weavingspace.tile_unit.TileUnit` setup
  methods in `weavingspace.tiling_geometries` use this method, where we are
  confident the returned dual is valid.

  Args:
    unit (TileUnit): the tiling for which the dual is required.

  Returns:
    gpd.GeoDataFrame: GeoDataFrame that could be the tiles attribute for
      a TileUnit of the dual tiling.

  """
  # get a local patch of this Tiling
  local_patch:gpd.GeoDataFrame = unit.get_local_patch(r = 1, include_0 = True)
  # Find the interior points of these tiles - these will be guaranteed
  # to have a sequence of surrounding tiles incident on them
  interior_pts = _get_interior_vertices(local_patch)
  # Compile a list of the polygons incident on the interior points
  cycles = [{poly_id for poly_id, p in enumerate(local_patch.geometry)
             if pt.distance(p) < RESOLUTION * 2} for pt in interior_pts]
  dual_faces = []
  for cycle in cycles:
    ids, pts = [], []
    for poly_id in cycle:
      ids.append(local_patch.tile_id[poly_id])
      poly = local_patch.geometry[poly_id]
      pts.append(get_incentre(poly))
    # sort them into CW order so they are well formed
    sorted_coords = _sort_cw([(pt.x, pt.y, ID)
                              for pt, ID in zip(pts, ids, strict = True)])
    dual_faces.append(
      (geom.Polygon([(pt_id[0], pt_id[1]) for pt_id in sorted_coords]),
       "".join([pt_id[2] for pt_id in sorted_coords])))
  # ensure the resulting face centroids are inside the original tile
  # displaced a little to avoid uncertainties at corners/edges
  # This is a guess at something that _could_ work, which turned out well!
  dual_faces = [(f, ID) for f, ID in dual_faces
                if affine.translate(
                  unit.get_prototile_from_vectors().loc[0, "geometry"],
                  RESOLUTION * 10, RESOLUTION * 10).contains(f.centroid)]
  gdf = gpd.GeoDataFrame(
    data = {"tile_id": [f[1] for f in dual_faces]}, crs = unit.crs,
    geometry = gpd.GeoSeries([f[0] for f in dual_faces]))
  # ensure no duplicates
  gdf = gdf.dissolve(by = "tile_id", as_index = False).explode(
    index_parts = False, ignore_index = True)
  gdf.tile_id = _relabel(gdf.tile_id)
  return gdf


def _relabel(data:Iterable) -> list:
  """Return supplied data reassigned with unique values from ascii_letters.

  Args:
    data (Iterable): the data to relabel

  Returns:
    list: the reassigned data

  """
  new_data = {}
  d_count = 0
  for d in data:
    if d not in new_data:
      new_data[d] = string.ascii_letters[d_count]
      d_count = d_count + 1
  return [new_data[d] for d in data]


def _sort_cw(
  pts_ids:list[tuple[float, float, str]]) -> list[tuple[float, float, str]]:
  """Sort supplied x,y,ID tuple into CW order relative to their mean centre.

  Args:
    pts_ids (list[tuple[float, float, str]]): A tuple of a pair of
      floats and a string.

  Returns:
    list: a list in the same format as supplied sorted into
      clockwise order of the point locations.

  """
  x = [p[0] for p in pts_ids]
  y = [p[1] for p in pts_ids]
  cx, cy = np.mean(x), np.mean(y)
  dx = [_ - cx for _ in x]
  dy = [_ - cy for _ in y]
  angles = [np.arctan2(dy, dx) for dx, dy in zip(dx, dy, strict = True)]
  d = dict(zip(angles, pts_ids, strict = True))
  return [pt_id for angle, pt_id in sorted(d.items(), reverse = True)]


def write_map_to_layers(gdf:gpd.GeoDataFrame,
                        fname:str = "output.gpkg",
                        tile_var:str = "tile_id") -> None:
  """Write supplied GeoDataFrame to a GPKG file.

  Geopackage file layers will be based on the `tile_var` attribute.

  Args:
    gdf (gpd.GeoDataFrame): the GeoDataFrame.
    fname (str, optional): filename to write.
    tile_var (str, optional): the attribute to use to separate
      output file into layers. Defaults to "tile_id".

  """
  grouped = gdf.groupby(tile_var, as_index = False)
  for e in pd.Series.unique(gdf[tile_var]):
    grouped.get_group(e).to_file(fname, layer = e, driver = "GPKG")


def get_collapse_distance(geometry:geom.Polygon) -> float:
  """Return distance under which polygon will disappear if negatively buffered.

  Performs a binary search between an upper bound based on the radius of circle
  of equal area to the polygon, and 0.

  Args:
    geometry (geom.Polygon): the polygon.

  Returns:
    float: its collapse distance.

  """
  if is_regular_polygon(geometry):
    return get_apothem(geometry)
  radius = np.sqrt(geometry.area / np.pi)
  lower = 0
  upper = radius
  delta = 1e12
  stop_delta = radius / 1000
  while delta > stop_delta:
    new_r = (lower + upper) / 2
    if geometry.buffer(-new_r).area > 0:
      lower = new_r
      delta = upper - new_r
    else:
      upper = new_r
      delta = new_r - lower
  return new_r


def get_largest_polygon(polygons:gpd.GeoSeries) -> gpd.GeoSeries:
  """Return largest polygon in a GeoSeries as a GeoSeries of one polygon.

  NOT CURRENTLY USED ANYWHERE IN MODULE.

  Args:
    polygons (gpd.GeoSeries): the set of polygons to pick from.

  Returns:
    gpd.GeoSeries: the largest polygon.

  """
  actual_polygons = [p for p in polygons.geometry
                     if isinstance(p, (geom.Polygon, geom.MultiPolygon))]
  areas = sorted(actual_polygons, key = lambda x: x.area, reverse = True)
  return gpd.GeoSeries(areas[:1])


def touch_along_an_edge(p1:geom.Polygon, p2:geom.Polygon) -> bool:
  """Test if two polygons touch along an edge.

  Checks that the intersection area of the two polygons buffered by
  a small amount is large enough to indicate that they neighbour at more
  than a corner.

  Args:
    p1 (geom.Polygon): First polygon
    p2 (geom.Polygon): Second polygon

  Returns:
    bool: True if they neighbour along an edge

  """
  return (
    p1.buffer(RESOLUTION, join_style = "mitre", cap_style = "square")
      .intersection(p2.buffer(RESOLUTION, join_style = "mitre",
                                          cap_style = "square"))
      .area > 16 * RESOLUTION ** 2)


def get_width_height_left_bottom(
    gs:gpd.GeoSeries,
  ) -> tuple[float,float,float,float]:
  """Return width, height, left and bottom limits of a GeoSeries.

  Args:
    gs (geopandas.GeoSeries): GeoSeries for which limits are required.

  Returns:
    tuple: four float values of width, height, left and bottom of gs.

  """
  bb = gs.total_bounds
  return (bb[2] - bb[0], bb[3] - bb[1], bb[0], bb[1])


def get_bounding_ellipse(
    shapes:gpd.GeoSeries,
    mag:float = 1.0,
  ) -> gpd.GeoSeries:
  """Return an ellipse containing the supplied shapes.

  The method used is to calculate the size of square that would contain
  the shapes, if they had an aspect ratio 1, then stretch the circle in
  the x, y directions according to the actual aspect ratio of the shapes.

  Args:
    shapes (gpd.GeoSeries): the shapes to be contained.
    mag (float, optional): optionally increase the size of the returned
      ellipse by this scale factor. Defaults to 1.0.

  Returns:
    gpd.GeoSeries: the set of shapes.

  """
  width, height, left, bottom = get_width_height_left_bottom(shapes)
  c = geom.Point(left + width / 2, bottom + height / 2)
  r = min(width, height) * np.sqrt(2)
  circle = [c.buffer(r)]
  return gpd.GeoSeries(circle, crs = shapes.crs).scale(
    width / r * mag / np.sqrt(2), height / r * mag / np.sqrt(2), origin = c)


def get_tiling_edges(tiles:gpd.GeoSeries) -> gpd.GeoSeries:
  """Return linestring GeoSeries from supplied polygon GeoSeries.

  This is used to allow display of edges of tiles in legend when they are
  masked by an ellipse (if we instead clip polygons then the ellipse edge
  will also show in the result.)

  Args:
    tiles (gpd.GeoSeries): Polygons to convert.

  Returns:
    gpd.GeoSeries: LineStrings from the supplied Polygons.

  """
  tiles = tiles.explode(ignore_index = True)
  edges = [geom.LineString(p.exterior.coords) for p in tiles]
  return gpd.GeoSeries(edges, crs = tiles.crs)


def get_polygon_sector(
    shape:geom.Polygon,
    start:float = 0.0,
    end:float = 1.0,
  ) -> geom.Polygon:
  """Return a sector of the provided Polygon.

  The returned sector is a section of the polygon boundary between the
  normalized start and end positions, and including the polygon centroid.
  Should (probably) only be applied to convex polygons.

  Args:
    shape (geom.Polygon): the Polygon.
    start (float): normalized start position along the boundary. Defaults to
      0.
    end (float): normalized start position along the boundary. Defaults to
      1.

  Returns:
    geom.Polygon: the requested polygon sector.

  """
  if start == end:
    # must return a null polygon since the calling context
    # expects to get something back... which most likely
    # is needed to align with other data
    return geom.Polygon()
  if start * end < 0:
    # either side of 0/1 so assume required sector includes 0
    e1, e2 = min(start, end), max(start, end)
    arc1 = shapely.ops.substring(geom.LineString(shape.exterior.coords),
                                 np.mod(e1, 1), 1, normalized = True)
    arc2 = shapely.ops.substring(geom.LineString(shape.exterior.coords),
                                 0, e2, normalized = True)
    sector = geom.Polygon(
      [shape.centroid] + list(arc1.coords) + list(arc2.coords)[1:])
  else:
    arc = shapely.ops.substring(geom.LineString(shape.exterior.coords),
                                start, end, normalized = True)
    sector = geom.Polygon([shape.centroid, *list(arc.coords)])
  return sector


def repair_polygon(
    polygon:geom.Polygon|geom.MultiPolygon|gpd.GeoSeries,
    shrink_then_grow:bool = True,
  ) -> geom.Polygon|gpd.GeoSeries:
  """Repair a polyon or GeoSeries by double buffering.

  This is method is often unofficially recommended (on stackexchange etc.)
  even in the shapely docs, to resolve topology issues and extraneous
  additional vertices appearing when spatial operations are repeatedly
  applied.

  Optionally the buffer may be applied in the opposite order (i.e. grow then
  shrink). This operation may also convert a MultiPolygon that has some 'stray'
  parts to a Polygon.

  Args:
    polygon(geom.Polygon|gpd.GeoSeries): Polygon or GeoSeries to clean.
    res (float, optional): buffer size to use. Defaults to 1e-3.
    shrink_then_grow (bool, optional): if True the negative buffer is
      applied first, otherwise the buffer operations are applied in
      reverse. Defaults to True.

  Returns:
    geom.Polygon|gpd.GeoSeries: the cleaned Polygon or GeoSeries.

  """
  if type(polygon) is gpd.GeoSeries:
    return gpd.GeoSeries([repair_polygon(p, shrink_then_grow)
                          for p in polygon]).set_crs(polygon.crs)
  if polygon is None:
    return None
  if shrink_then_grow:
    return polygon.buffer(
      -RESOLUTION * 10, join_style = "mitre", cap_style = "square").buffer(
       RESOLUTION * 10, join_style = "mitre", cap_style = "square")
  return polygon.buffer(
     RESOLUTION * 10, join_style = "mitre", cap_style = "square").buffer(
    -RESOLUTION * 10, join_style = "mitre", cap_style = "square")


def safe_union(
    gs:gpd.GeoSeries,
    as_polygon:bool = False,
  ) -> gpd.GeoSeries|geom.Polygon:
  """Union supplied GeoSeries buffering to avoid gaps and internal edges.

  Optionally returns a Polygon or a GeoSeries.

  NOTE: APPEARS NOT TO BE VERY SAFE... find workarounds for the few occasions
  on which it is used.

  Frequently when unioning polygons that are ostensibly adjacent 'rogue'
  internal boundaries remain in the result. We can avoid this by buffering the
  polygons before unioning them, then reversing the buffer on the unioned
  shape.

  Args:
    gs (gpd.GeoSeries): the Polygons to union.
    res (float, optional): size of the buffer to use. Defaults to 1e-3.
    as_polygon (bool, optional): if True returns a Polygon, otherwise
      returns a one Polygon GeoSeries. Defaults to False.

  Returns:
    gpd.GeoSeries|geom.Polygon: the resulting union of supplied
      polygons.

  """
  r = RESOLUTION * 10
  union = gpd.GeoSeries(
    [p.buffer(r, join_style = "mitre", cap_style = "square") for p in gs],
    crs = gs.crs).union_all().buffer(-r, join_style = "mitre",
                                         cap_style = "square")
  if as_polygon:
    return union
  return gpd.GeoSeries([union], crs = gs.crs)


def get_translation_transform(dx:float, dy:float) -> tuple[float,...]:
  """Return shapely affine transform tuple for a translation.

  Args:
      dx (float): translation distance in x direction.
      dy (float): translation distance in y direction.

  Returns:
    list[float]: a six item list of floats, per the shapely.affinity.
    affine_transform method, see
      https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations

  """
  return (1, 0, 0, 1, dx, dy)


def get_rotation_transform(
    angle:float, centre:tuple[float,...]|None = None) -> tuple[float,...]:
  """Return the shapely affine transform tuple for a rotation.

  Rotation is optionally about a supplied centre point.

  Args:
    angle (float): the angle of rotation (in degrees).
    centre (tuple[float], optional): An option centre location. Defaults to
    None, which will in turn be converted to (0, 0).

  Returns:
    list[float]: a six item list of floats, per the shapely.affinity.
      affine_transform method, see
        https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations

  """
  if centre is None or np.allclose((0, 0), centre, atol = RESOLUTION):
    a = np.radians(angle)
    return (np.cos(a), -np.sin(a), np.sin(a), np.cos(a), 0, 0)
  dx, dy = centre
  t1 = get_translation_transform(-dx, -dy)
  r = get_rotation_transform(angle)
  t2 = get_translation_transform(dx, dy)
  return combine_transforms([t1, r, t2])


def get_reflection_transform(
    angle:float, centre:tuple[float,...]|None = None) -> tuple[float,...]:
  """Return a shapely affine reflection transform.

  Reflection is in a line at the specified angle, optionally through a
  specified centre point.

  Args:
    angle (float): angle to the x-axis of the line of reflection.
    centre (tuple[float], optional): point through which the line of
      reflection passes. Defaults to None, which
      will in turn be converted to (0, 0).

  Returns:
    list[float]: a six item list of floats, per the shapely.affinity.
      affine_transform method, see
        https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations

  """
  if centre is None or np.allclose((0, 0), centre, atol = RESOLUTION):
    A = 2 * np.radians(angle)
    return (np.cos(A), np.sin(A), np.sin(A), -np.cos(A), 0, 0)
  dx, dy = centre
  t1 = get_translation_transform(-dx, -dy)
  r = get_reflection_transform(angle)
  t2 = get_translation_transform(dx, dy)
  return combine_transforms([t1, r, t2])


def combine_transforms(transforms:list[tuple[float,...]]) -> tuple[float,...]:
  """Return shapely affine transform that combines supplied list of transforms.

  Transforms are applied in the order of the supplied list

  Args:
    transforms (list[list[float]]): sequence of transforms to combine.

  Returns:
    list[float]: a transform tuple combining the supplied transforms applied
      in order, see
        https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations

  """
  result = np.identity(3)
  for t in transforms:
    result = as_numpy_matrix(t) @ result
  return as_shapely_transform(result)


def reverse_transform(transform:tuple[float,...]) -> tuple[float,...]:
  """Return inverse shapely affine transform of the supplied transform.

  Args:
    transform (list[float]): the transform for which the inverse is desired.

  Returns:
    list[float]: shapely affine transform tuple that will invert the supplied
      transform.

  """
  return as_shapely_transform(
    np.linalg.inv(as_numpy_matrix(transform)))


def as_shapely_transform(arr:np.ndarray) -> tuple[float,...]:
  """Return shapely affine transform equivalent to supplied numpy matrix.

  The numpy matrix is a conventional augmented (i.e. 3x3) affine transform
  matrix.

  Args:
    arr (np.array): augmented affine transform matrix of the desired
      transform.

  Returns:
    list[float]: desired shapely affine transform list of floats.

  """
  return (arr[0][0], arr[0][1], arr[1][0], arr[1][1], arr[0][2], arr[1][2])


def as_numpy_matrix(transform:tuple[float,...]) -> np.ndarray:
  """Convert shapely affine transform to augmented affine transform matrix.

  Args:
    transform (list[float]): the transform in shapely format.

  Returns:
    np.array: the transform in numpy matrix format.

  """
  return np.array([[transform[0], transform[1], transform[4]],
                   [transform[2], transform[3], transform[5]],
                   [           0,            0,            1]])


class StraightLine(NamedTuple):
  """Straight line in canonical Ax+By+C=0 form."""

  A: float
  B: float
  C: float


def get_straight_line(
    p1:geom.Point, p2:geom.Point,
    perpendicular:bool = False,
  ) -> StraightLine:
  """Return StraightLine p1-p2, optionally perpendicular through the midpoint.

  Args:
    p1 (geom.Point): First point.
    p2 (geom.Point): Second point
    perpendicular (bool, optional): If True the perpendicular bisector is
      returned. Defaults to False.

  Returns:
    StraightLine: _description_

  """
  if perpendicular:
    ls = affine.rotate(geom.LineString([p1, p2]), 90)
    pts = list(ls.coords)
    x1, y1 = pts[0]
    x2, y2 = pts[1]
  else:
    x1, y1 = p1.x, p1.y
    x2, y2 = p2.x, p2.y
  return StraightLine(y1 - y2, x2 - x1, x1 * y2 - x2 * y1)


def get_intersection(line1:StraightLine,
                     line2:StraightLine) -> geom.Point|None:
  """Return point of intersection of straight lines.

  Args:
    line1 (StraightLine): First straight line.
    line2 (StraightLine): Second straight line.

  Returns:
    geom.Point|None: Point is returned if the lines intersect, otherwise None.

  """
  x_set, y_set = False, False
  denominator = line1.A * line2.B - line2.A * line1.B
  if np.isclose(line1.A, 0, atol = 1e-4, rtol = 1e-4):
    y = -line1.C / line1.B
    y_set = True
  elif np.isclose(line2.A, 0, atol = 1e-4, rtol = 1e-4):
    y = -line2.C / line2.B
    y_set = True
  if np.isclose(line1.B, 0, atol = 1e-4, rtol = 1e-4):
    x = -line1.C / line1.A
    x_set = True
  elif np.isclose(line2.B, 0, atol = 1e-4, rtol = 1e-4):
    x = -line2.C / line2.A
    x_set = True
  if np.isclose(denominator, 0, atol = 1e-4, rtol = 1e-4):
    return None
  x = x if x_set else (line1.B * line2.C - line2.B * line1.C) / denominator
  y = y if y_set else (line1.C * line2.A - line2.C * line1.A) / denominator
  return geom.Point(x, y)
